import ast
import os.path
from inspect import isclass
from typing import Dict, List, Optional, Tuple

from boa3 import constants
from boa3.analyser.astanalyser import IAstAnalyser
from boa3.compiler.codegenerator.codegenerator import CodeGenerator
from boa3.compiler.codegenerator.generatordata import GeneratorData
from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping
from boa3.constants import SYS_VERSION_INFO
from boa3.model.builtin.builtin import Builtin
from boa3.model.builtin.interop.interop import Interop
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.imports.package import Package
from boa3.model.method import Method
from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.binaryop import BinaryOp
from boa3.model.operation.operation import IOperation
from boa3.model.operation.unary.unaryoperation import UnaryOperation
from boa3.model.property import Property
from boa3.model.symbol import ISymbol
from boa3.model.type.classes.classtype import ClassType
from boa3.model.type.classes.userclass import UserClass
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.type import IType, Type
from boa3.model.variable import Variable


class VisitorCodeGenerator(IAstAnalyser):
    """
    This class is responsible for walk through the ast.

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.

    :ivar generator:
    """

    def __init__(self, generator: CodeGenerator, filename: str = None):
        super().__init__(ast.parse(""), filename=filename, log=True)

        self.generator = generator
        self.current_method: Optional[Method] = None
        self.current_class: Optional[UserClass] = None
        self.symbols = generator.symbol_table

        self.global_stmts: List[ast.AST] = []
        self._is_generating_initialize = False
        self._root_module: ast.AST = self._tree

    @property
    def _symbols(self) -> Dict[str, ISymbol]:
        symbol_table = self.symbols.copy()

        if isinstance(self.current_class, UserClass):
            symbol_table.update(self.current_class.symbols)

        return symbol_table

    def include_instruction(self, node: ast.AST, address: int):
        if self.current_method is not None and address in VMCodeMapping.instance().code_map:
            bytecode = VMCodeMapping.instance().code_map[address]
            from boa3.model.debuginstruction import DebugInstruction
            self.current_method.include_instruction(DebugInstruction.build(node, bytecode))

    def build_data(self, origin_node: Optional[ast.AST],
                   symbol_id: Optional[str] = None,
                   symbol: Optional[ISymbol] = None,
                   result_type: Optional[IType] = None,
                   index: Optional[int] = None,
                   already_generated: bool = False) -> GeneratorData:

        if isinstance(symbol, IType) and result_type is None:
            result_type = symbol
            symbol = None

        if symbol is None and symbol_id is not None:
            # try to find the symbol if the id is known
            if symbol_id in self._symbols:
                found_symbol = self._symbols[symbol_id]
            else:
                _, found_symbol = self.generator.get_symbol(symbol_id)
                if found_symbol is Type.none:
                    # generator get_symbol returns Type.none if the symbol is not found
                    found_symbol = None

            if found_symbol is not None:
                symbol = found_symbol

        return GeneratorData(origin_node, symbol_id, symbol, result_type, index, already_generated)

    def visit(self, node: ast.AST) -> GeneratorData:
        result = super().visit(node)
        if not isinstance(result, GeneratorData):
            result = self.build_data(node)
        return result

    def visit_to_map(self, node: ast.AST, generate: bool = False) -> GeneratorData:
        address: int = VMCodeMapping.instance().bytecode_size
        if isinstance(node, ast.Expr):
            value = self.visit_Expr(node, generate)
        elif generate:
            value = self.visit_to_generate(node)
        else:
            value = self.visit(node)

        if not isinstance(node, (ast.For, ast.While, ast.If)):
            # control flow nodes must map each of their instructions
            self.include_instruction(node, address)
        return value

    def visit_to_generate(self, node) -> GeneratorData:
        """
        Visitor to generate the nodes that the primary visitor is used to retrieve value

        :param node: an ast node
        """
        if isinstance(node, ast.AST):
            result = self.visit(node)

            if not result.already_generated and result.symbol_id is not None:
                if self.is_exception_name(result.symbol_id):
                    self.generator.convert_new_exception()
                else:
                    # TODO: validate function calls
                    is_internal = hasattr(node, 'is_internal_call') and node.is_internal_call
                    class_type = result.type

                    if (self.is_implemented_class_type(result.type)
                            and len(result.symbol_id.split(constants.ATTRIBUTE_NAME_SEPARATOR)) > 1):
                        # if the symbol id has the attribute separator and the top item on the stack is a user class,
                        # then this value is an attribute from that class
                        # change the id for correct generation
                        result.symbol_id = result.symbol_id.split(constants.ATTRIBUTE_NAME_SEPARATOR)[-1]

                    elif isinstance(result.index, Package):
                        class_type = None

                    self.generator.convert_load_symbol(result.symbol_id, is_internal=is_internal, class_type=class_type)

                result.already_generated = True

            return result
        else:
            index = self.generator.convert_literal(node)
            return self.build_data(node, index=index)

    def is_exception_name(self, exc_id: str) -> bool:
        global_symbols = globals()
        if exc_id in global_symbols or exc_id in global_symbols['__builtins__']:
            symbol = (global_symbols[exc_id]
                      if exc_id in global_symbols
                      else global_symbols['__builtins__'][exc_id])
            if isclass(symbol) and issubclass(symbol, BaseException):
                return True
        return False

    def _remove_inserted_opcodes_since(self, last_address: int, last_stack_size: Optional[int] = None):
        if VMCodeMapping.instance().bytecode_size > last_address:
            # remove opcodes inserted during the evaluation of the symbol
            VMCodeMapping.instance().remove_opcodes(last_address, VMCodeMapping.instance().bytecode_size)

        if isinstance(last_stack_size, int) and last_stack_size < self.generator.stack_size:
            # remove any additional values pushed to the stack during the evalution of the symbol
            for _ in range(self.generator.stack_size - last_stack_size):
                self.generator._stack_pop()

    def _get_unique_name(self, name_id: str, node: ast.AST) -> str:
        return '{0}{2}{1}'.format(node.__hash__(), name_id, constants.VARIABLE_NAME_SEPARATOR)

    def set_filename(self, filename: str):
        if isinstance(filename, str) and os.path.isfile(filename):
            if constants.PATH_SEPARATOR != os.path.sep:
                self.filename = filename.replace(constants.PATH_SEPARATOR, os.path.sep)
            else:
                self.filename = filename

    def visit_Module(self, module: ast.Module) -> GeneratorData:
        """
        Visitor of the module node

        Fills module symbol table

        :param module:
        """
        global_stmts = [node for node in module.body if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))]
        function_stmts = module.body[len(global_stmts):]
        mandatory_global_stmts = []
        for stmt in global_stmts:
            if isinstance(stmt, ast.ClassDef):
                class_symbol = self.get_symbol(stmt.name)
                if isinstance(class_symbol, UserClass) and len(class_symbol.class_variables) > 0:
                    mandatory_global_stmts.append(stmt)
            elif not isinstance(stmt, (ast.Import, ast.ImportFrom)):
                mandatory_global_stmts.append(stmt)

        for stmt in function_stmts:
            self.visit(stmt)

        if self.generator.initialize_static_fields():
            last_symbols = self.symbols  # save to revert in the end and not compromise consequent visits
            class_non_static_stmts = []

            for node in global_stmts.copy():
                if isinstance(node, ast.ClassDef):
                    class_origin_module = None
                    if hasattr(node, 'origin'):
                        class_origin_module = node.origin
                        if (node.origin is not self._root_module
                                and hasattr(node.origin, 'symbols')):
                            # symbols unique to imports are not included in the symbols
                            self.symbols = node.origin.symbols

                    class_variables = []
                    class_functions = []
                    for stmt in node.body:
                        if isinstance(stmt, (ast.Assign, ast.AugAssign, ast.AnnAssign)):
                            class_variables.append(stmt)
                        else:
                            class_functions.append(stmt)

                    if len(class_functions) > 0:
                        if class_origin_module is not None and not hasattr(node, 'origin'):
                            node.origin = class_origin_module

                        cls_fun = node
                        if len(class_variables) > 0:
                            cls_var = node
                            cls_fun = self.clone(node)
                            cls_fun.body = class_functions
                            cls_var.body = class_variables
                        else:
                            global_stmts.remove(cls_fun)

                        class_non_static_stmts.append(cls_fun)

            # to generate the 'initialize' method for Neo
            self._log_info(f"Compiling '{constants.INITIALIZE_METHOD_ID}' function")
            self._is_generating_initialize = True
            for stmt in global_stmts:
                cur_tree = self._tree
                cur_filename = self.filename
                if hasattr(stmt, 'origin'):
                    if hasattr(stmt.origin, 'filename'):
                        self.set_filename(stmt.origin.filename)
                    self._tree = stmt.origin

                self.visit(stmt)
                self.filename = cur_filename
                self._tree = cur_tree

            self._is_generating_initialize = False
            self.generator.end_initialize()

            # generate any symbol inside classes that's not variables AFTER generating 'initialize' method
            for stmt in class_non_static_stmts:
                cur_tree = self._tree
                if hasattr(stmt, 'origin'):
                    self._tree = stmt.origin
                self.visit(stmt)
                self._tree = cur_tree

            self.symbols = last_symbols  # revert in the end to not compromise consequent visits
            self.generator.additional_symbols = None

        elif len(function_stmts) > 0 or len(mandatory_global_stmts) > 0:
            # to organize syntax tree nodes from other modules
            for stmt in global_stmts:
                if not hasattr(stmt, 'origin'):
                    stmt.origin = module

            module.symbols = self._symbols
            self.global_stmts.extend(global_stmts)
        else:
            # to generate objects when there are no static variables to generate 'initialize'
            for stmt in global_stmts:
                self.visit(stmt)

        return self.build_data(module)

    def visit_ClassDef(self, node: ast.ClassDef) -> GeneratorData:
        if node.name in self.symbols:
            class_symbol = self.symbols[node.name]
            if isinstance(class_symbol, UserClass):
                self.current_class = class_symbol

            if self._is_generating_initialize:
                address = self.generator.bytecode_size
                self.generator.convert_new_empty_array(len(class_symbol.class_variables), class_symbol)
                self.generator.convert_store_variable(node.name, address)
            else:
                init_method = class_symbol.constructor_method()
                if isinstance(init_method, Method) and init_method.start_address is None:
                    self.generator.generate_implicit_init_user_class(init_method)

        for stmt in node.body:
            self.visit(stmt)

        self.current_class = None
        return self.build_data(node)

    def visit_FunctionDef(self, function: ast.FunctionDef) -> GeneratorData:
        """
        Visitor of the function definition node

        Generates the Neo VM code for the function

        :param function: the python ast function definition node
        """
        method = self._symbols[function.name]

        if isinstance(method, Property):
            method = method.getter

        if isinstance(method, Method):
            self.current_method = method
            if isinstance(self.current_class, ClassType):
                function_name = self.current_class.identifier + constants.ATTRIBUTE_NAME_SEPARATOR + function.name
            else:
                function_name = function.name

            self._log_info(f"Compiling '{function_name}' function")

            if method.is_public or method.is_called:
                if not isinstance(self.current_class, ClassType) or not self.current_class.is_interface:
                    self.generator.convert_begin_method(method)

                    for stmt in function.body:
                        self.visit_to_map(stmt)

                    self.generator.convert_end_method(function.name)

            self.current_method = None

        return self.build_data(function, symbol=method, symbol_id=function.name)

    def visit_Return(self, ret: ast.Return) -> GeneratorData:
        """
        Visitor of a function return node

        :param ret: the python ast return node
        """
        if self.generator.stack_size > 0:
            self.generator.clear_stack(True)
        if self.current_method.return_type is not Type.none:
            result = self.visit_to_generate(ret.value)
            if result.type is Type.none and not self.generator.is_none_inserted():
                self.generator.convert_literal(None)
        self.generator.insert_return()

        return self.build_data(ret)

    def store_variable(self, *var_ids: Tuple[str, Optional[ast.AST], int], value: ast.AST):
        # if the value is None, it is a variable declaration
        if value is not None:
            if len(var_ids) == 1:
                # it's a simple assignment
                var_id, index, address = var_ids[0]
                if index is None:
                    # if index is None, then it is a variable assignment
                    result_data = self.visit_to_generate(value)

                    if result_data.type is Type.none and not self.generator.is_none_inserted():
                        self.generator.convert_literal(None)
                    self.generator.convert_store_variable(var_id, address, self.current_class)
                else:
                    # if not, it is an array assignment
                    self.generator.convert_load_symbol(var_id)
                    self.visit_to_generate(index)

                    aux_index = VMCodeMapping.instance().bytecode_size
                    value_data = self.visit_to_generate(value)
                    value_address = value_data.index if value_data.index is not None else aux_index

                    self.generator.convert_set_item(value_address)

            elif len(var_ids) > 0:
                # it's a chained assignment
                self.visit_to_generate(value)
                for pos, (var_id, index, address) in enumerate(reversed(var_ids)):
                    if index is None:
                        # if index is None, then it is a variable assignment
                        if pos < len(var_ids) - 1:
                            self.generator.duplicate_stack_top_item()
                        self.generator.convert_store_variable(var_id, address)
                    else:
                        # if not, it is an array assignment
                        if pos < len(var_ids) - 1:
                            self.generator.convert_load_symbol(var_id)
                            self.visit_to_generate(index)
                            fix_index = VMCodeMapping.instance().bytecode_size
                            self.generator.duplicate_stack_item(3)
                        else:
                            self.visit_to_generate(index)
                            fix_index = VMCodeMapping.instance().bytecode_size
                            self.generator.convert_load_symbol(var_id)
                            self.generator.swap_reverse_stack_items(3)
                        self.generator.convert_set_item(fix_index)

    def visit_AnnAssign(self, ann_assign: ast.AnnAssign) -> GeneratorData:
        """
        Visitor of an annotated assignment node

        :param ann_assign: the python ast variable assignment node
        """
        var_value_address = self.generator.bytecode_size
        var_data = self.visit(ann_assign.target)
        var_id = var_data.symbol_id
        # filter to find the imported variables
        if (var_id not in self.generator.symbol_table
                and hasattr(ann_assign, 'origin')
                and isinstance(ann_assign.origin, ast.AST)):
            var_id = self._get_unique_name(var_id, ann_assign.origin)
        self.store_variable((var_id, None, var_value_address), value=ann_assign.value)
        return self.build_data(ann_assign)

    def visit_Assign(self, assign: ast.Assign) -> GeneratorData:
        """
        Visitor of an assignment node

        :param assign: the python ast variable assignment node
        """
        vars_ids: List[Tuple[str, Optional[ast.AST], int]] = []
        for target in assign.targets:
            var_value_address = self.generator.bytecode_size
            target_data: GeneratorData = self.visit(target)

            var_id = target_data.symbol_id
            var_index = target_data.index

            # filter to find the imported variables
            if (var_id not in self.generator.symbol_table
                    and hasattr(assign, 'origin')
                    and isinstance(assign.origin, ast.AST)):
                var_id = self._get_unique_name(var_id, assign.origin)
            vars_ids.append((var_id, var_index, var_value_address))

        self.store_variable(*vars_ids, value=assign.value)
        return self.build_data(assign)

    def visit_AugAssign(self, aug_assign: ast.AugAssign) -> GeneratorData:
        """
        Visitor of an augmented assignment node

        :param aug_assign: the python ast augmented assignment node
        """
        var_data = self.visit(aug_assign.target)
        var_id = var_data.symbol_id
        # filter to find the imported variables
        if (var_id not in self.generator.symbol_table
                and hasattr(aug_assign, 'origin')
                and isinstance(aug_assign.origin, ast.AST)):
            var_id = self._get_unique_name(var_id, aug_assign.origin)

        if isinstance(var_data.type, UserClass):
            self.generator.duplicate_stack_top_item()

        self.generator.convert_load_symbol(var_id)
        value_address = self.generator.bytecode_size
        self.visit_to_generate(aug_assign.value)
        self.generator.convert_operation(aug_assign.op)
        self.generator.convert_store_variable(var_id, value_address)
        return self.build_data(aug_assign)

    def visit_Subscript(self, subscript: ast.Subscript) -> GeneratorData:
        """
        Visitor of a subscript node

        :param subscript: the python ast subscript node
        """
        if isinstance(subscript.slice, ast.Slice):
            return self.visit_Subscript_Slice(subscript)

        return self.visit_Subscript_Index(subscript)

    def visit_Subscript_Index(self, subscript: ast.Subscript) -> GeneratorData:
        """
        Visitor of a subscript node with index

        :param subscript: the python ast subscript node
        """
        index = None
        symbol_id = None
        if isinstance(subscript.ctx, ast.Load):
            # get item
            value_data = self.visit_to_generate(subscript.value)
            slice = subscript.slice.value if isinstance(subscript.slice, ast.Index) else subscript.slice
            self.visit_to_generate(slice)
            self.generator.convert_get_item()

            value_type = value_data.type
        else:
            # set item
            var_data = self.visit(subscript.value)

            index = subscript.slice.value if isinstance(subscript.slice, ast.Index) else subscript.slice
            symbol_id = var_data.symbol_id
            value_type = var_data.type

        result_type = value_type.item_type if isinstance(value_type, SequenceType) else value_type
        return self.build_data(subscript, result_type=result_type,
                               symbol_id=symbol_id, index=index)

    def visit_Subscript_Slice(self, subscript: ast.Subscript) -> GeneratorData:
        """
        Visitor of a subscript node with slice

        :param subscript: the python ast subscript node
        """
        lower_omitted = subscript.slice.lower is None
        upper_omitted = subscript.slice.upper is None
        step_omitted = subscript.slice.step is None

        self.visit_to_generate(subscript.value)

        step_negative = True if not step_omitted and subscript.slice.step.n < 0 else False
        # if step is negative, then consider the value reversed
        if step_negative:
            self.generator.convert_array_negative_stride()

        # if both are explicit
        if not lower_omitted and not upper_omitted:
            addresses = [VMCodeMapping.instance().bytecode_size]
            self.visit_to_generate(subscript.slice.lower)

            # length of slice
            addresses.append(VMCodeMapping.instance().bytecode_size)
            self.visit_to_generate(subscript.slice.upper)

            self.generator.convert_get_sub_array(addresses, step_negative)
        # only one of them is omitted
        elif lower_omitted != upper_omitted:
            # start position is omitted
            if lower_omitted:
                self.visit_to_generate(subscript.slice.upper)
                self.generator.convert_get_array_beginning(step_negative)
            # end position is omitted
            else:
                self.generator.duplicate_stack_top_item()
                # length of slice
                self.generator.convert_builtin_method_call(Builtin.Len)
                self.visit_to_generate(subscript.slice.lower)
                self.generator.convert_get_array_ending(step_negative)
        else:
            self.generator.convert_copy()

        if not step_omitted:
            self.visit_to_generate(subscript.slice.step)
            self.generator.convert_get_stride()

        return self.build_data(subscript)

    def _convert_unary_operation(self, operand, op):
        self.visit_to_generate(operand)
        self.generator.convert_operation(op)

    def _convert_binary_operation(self, left, right, op):
        self.visit_to_generate(left)
        self.visit_to_generate(right)
        self.generator.convert_operation(op)

    def visit_BinOp(self, bin_op: ast.BinOp) -> GeneratorData:
        """
        Visitor of a binary operation node

        :param bin_op: the python ast binary operation node
        """
        if isinstance(bin_op.op, BinaryOperation):
            self._convert_binary_operation(bin_op.left, bin_op.right, bin_op.op)

        return self.build_data(bin_op)

    def visit_UnaryOp(self, un_op: ast.UnaryOp) -> GeneratorData:
        """
        Visitor of a binary operation node

        :param un_op: the python ast binary operation node
        """
        if isinstance(un_op.op, UnaryOperation):
            self._convert_unary_operation(un_op.operand, un_op.op)

        return self.build_data(un_op)

    def visit_Compare(self, compare: ast.Compare) -> GeneratorData:
        """
        Visitor of a compare operation node

        :param compare: the python ast compare operation node
        """
        operation_symbol: IOperation = BinaryOp.And
        converted: bool = False
        left = compare.left
        for index, op in enumerate(compare.ops):
            right = compare.comparators[index]
            if isinstance(op, IOperation):
                if isinstance(op, BinaryOperation):
                    self._convert_binary_operation(left, right, op)
                else:
                    operand = left
                    if isinstance(operand, ast.NameConstant) and operand.value is None:
                        operand = right
                    self._convert_unary_operation(operand, op)
                # if it's more than two comparators, must include AND between the operations
                if not converted:
                    converted = True
                    operation_symbol = op
                else:
                    operation_symbol = BinaryOp.And
                    self.generator.convert_operation(BinaryOp.And)
            left = right

        return self.build_data(compare, symbol=operation_symbol, result_type=operation_symbol.result)

    def visit_BoolOp(self, bool_op: ast.BoolOp) -> GeneratorData:
        """
        Visitor of a compare operation node

        :param bool_op: the python ast boolean operation node
        """
        if isinstance(bool_op.op, BinaryOperation):
            left = bool_op.values[0]
            self.visit_to_generate(left)
            for index, right in enumerate(bool_op.values[1:]):
                self.visit_to_generate(right)
                self.generator.convert_operation(bool_op.op)

        return self.build_data(bool_op)

    def visit_While(self, while_node: ast.While) -> GeneratorData:
        """
        Visitor of a while statement node

        :param while_node: the python ast while statement node
        """
        start_addr: int = self.generator.convert_begin_while()
        for stmt in while_node.body:
            self.visit_to_map(stmt, generate=True)

        test_address: int = VMCodeMapping.instance().bytecode_size
        test_data = self.visit_to_map(while_node.test, generate=True)
        self.generator.convert_end_while(start_addr, test_address)

        else_begin_address: int = self.generator.last_code_start_address
        for stmt in while_node.orelse:
            self.visit_to_map(stmt, generate=True)

        self.generator.convert_end_loop_else(start_addr, else_begin_address, len(while_node.orelse) > 0)
        return self.build_data(while_node, index=start_addr)

    def visit_For(self, for_node: ast.For) -> GeneratorData:
        """
        Visitor of for statement node

        :param for_node: the python ast for node
        """
        self.visit_to_generate(for_node.iter)
        start_address = self.generator.convert_begin_for()

        if isinstance(for_node.target, tuple):
            for target in for_node.target:
                var_data = self.visit_to_map(target)
                self.generator.convert_store_variable(var_data.symbol_id)
        else:
            var_data = self.visit(for_node.target)
            self.generator.convert_store_variable(var_data.symbol_id)

        for stmt in for_node.body:
            self.visit_to_map(stmt, generate=True)

        # TODO: remove when optimizing for generation
        if self.current_method is not None:
            self.current_method.remove_instruction(for_node.lineno, for_node.col_offset)

        condition_address = self.generator.convert_end_for(start_address)
        self.include_instruction(for_node, condition_address)
        else_begin = self.generator.last_code_start_address

        for stmt in for_node.orelse:
            self.visit_to_map(stmt, generate=True)

        self.generator.convert_end_loop_else(start_address,
                                             else_begin,
                                             has_else=len(for_node.orelse) > 0,
                                             is_for=True)
        return self.build_data(for_node)

    def visit_If(self, if_node: ast.If) -> GeneratorData:
        """
        Visitor of if statement node

        :param if_node: the python ast if statement node
        """
        test = self.visit_to_map(if_node.test, generate=True)

        if not Type.bool.is_type_of(test.type) and test.type is not None:
            self.generator.convert_builtin_method_call(Builtin.Bool)

        start_addr: int = self.generator.convert_begin_if()
        for stmt in if_node.body:
            self.visit_to_map(stmt, generate=True)

        ends_with_if = len(if_node.body) > 0 and isinstance(if_node.body[-1], ast.If)

        if len(if_node.orelse) > 0:
            start_addr = self.generator.convert_begin_else(start_addr, ends_with_if)
            for stmt in if_node.orelse:
                self.visit_to_map(stmt, generate=True)

        self.generator.convert_end_if(start_addr)
        return self.build_data(if_node)

    def visit_Expr(self, expr: ast.Expr, generate: bool = False) -> GeneratorData:
        """
        Visitor of an expression node

        :param expr: the python ast expression node
        :param generate: if it should convert the value
        """
        last_stack = self.generator.stack_size
        if generate:
            value = self.visit_to_generate(expr.value)
        else:
            value = self.visit(expr.value)

        new_stack = self.generator.stack_size
        for x in range(last_stack, new_stack):
            self.generator.remove_stack_top_item()

        return value

    def visit_IfExp(self, if_node: ast.IfExp) -> GeneratorData:
        """
        Visitor of if expression node

        :param if_node: the python ast if statement node
        """
        self.visit_to_map(if_node.test, generate=True)

        start_addr: int = self.generator.convert_begin_if()
        body_data = self.visit_to_map(if_node.body, generate=True)

        start_addr = self.generator.convert_begin_else(start_addr)
        else_data = self.visit_to_map(if_node.orelse, generate=True)

        self.generator.convert_end_if(start_addr)
        return self.build_data(if_node, result_type=Type.union.build([body_data.type, else_data.type]))

    def visit_Assert(self, assert_node: ast.Assert) -> GeneratorData:
        """
        Visitor of the assert node

        :param assert_node: the python ast assert node
        """
        self.visit_to_generate(assert_node.test)

        if assert_node.msg is not None:
            self.generator.duplicate_stack_top_item()
            self.generator.insert_not()

            # if assert is false, log the message
            start_addr: int = self.generator.convert_begin_if()

            self.visit_to_generate(assert_node.msg)
            self.generator.convert_builtin_method_call(Interop.Log)

            self.generator.convert_end_if(start_addr)

        self.generator.convert_assert()
        return self.build_data(assert_node)

    def visit_Call(self, call: ast.Call) -> GeneratorData:
        """
        Visitor of a function call node

        :param call: the python ast function call node
        :returns: The called function return type
        """
        # the parameters are included into the stack in the reversed order
        last_address = VMCodeMapping.instance().bytecode_size
        last_stack = self.generator.stack_size

        func_data = self.visit(call.func)
        function_id = func_data.symbol_id
        # if the symbol is not a method, check if it is a class method
        if (isinstance(func_data.type, ClassType) and func_data.symbol_id in func_data.type.symbols
                and isinstance(func_data.type.symbols[func_data.symbol_id], IExpression)):
            symbol = func_data.type.symbols[func_data.symbol_id]
        else:
            symbol = func_data.symbol

        if not isinstance(symbol, Method):
            is_internal = hasattr(call, 'is_internal_call') and call.is_internal_call
            _, symbol = self.generator.get_symbol(function_id, is_internal=is_internal)

        if self.is_implemented_class_type(symbol):
            self.generator.convert_init_user_class(symbol)
            symbol = symbol.constructor_method()
        args_addresses: List[int] = []

        has_cls_or_self_argument = isinstance(symbol, Method) and symbol.has_cls_or_self
        if not has_cls_or_self_argument:
            self._remove_inserted_opcodes_since(last_address, last_stack)

        if isinstance(symbol, Method):
            # self or cls is already generated
            call_args = (call.args[1:]
                         if has_cls_or_self_argument and len(call.args) == len(symbol.args)
                         else call.args)
            args_to_generate = [arg for index, arg in enumerate(call_args) if index in symbol.args_to_be_generated()]
            keywords_dict = {keyword.arg: keyword.value for keyword in call.keywords}
            keywords_with_index = {index: keywords_dict[arg_name]
                                   for index, arg_name in enumerate(symbol.args)
                                   if arg_name in keywords_dict}

            for index in keywords_with_index:
                if index < len(args_to_generate):
                    # override default values
                    args_to_generate[index] = keywords_with_index[index]
                else:
                    # put keywords
                    args_to_generate.append(keywords_with_index[index])

        else:
            args_to_generate = call.args

        if isinstance(symbol, IBuiltinMethod):
            reordered_args = []
            for index in symbol.generation_order:
                if 0 <= index < len(args_to_generate):
                    reordered_args.append(args_to_generate[index])

            args = reordered_args
        else:
            args = reversed(args_to_generate)

        args_begin_address = self.generator.last_code_start_address
        for arg in args:
            args_addresses.append(
                VMCodeMapping.instance().bytecode_size
            )
            self.visit_to_generate(arg)
        if has_cls_or_self_argument:
            num_args = len(args_addresses)
            if self.generator.stack_size > num_args:
                value = self.generator._stack_pop(-num_args - 1)
                self.generator._stack_append(value)
            end_address = VMCodeMapping.instance().move_to_end(last_address, args_begin_address)
            if not symbol.is_init:
                args_addresses.append(end_address)

        if self.is_exception_name(function_id):
            self.generator.convert_new_exception(len(call.args))
        elif isinstance(symbol, type(Builtin.Super)) and len(args_to_generate) == 0:
            self_or_cls_id = list(self.current_method.args)[0]
            self.generator.convert_load_symbol(self_or_cls_id)
        elif isinstance(symbol, IBuiltinMethod):
            self.generator.convert_builtin_method_call(symbol, args_addresses)
        else:
            self.generator.convert_load_symbol(function_id, args_addresses)

        return self.build_data(call, symbol=symbol, symbol_id=function_id,
                               result_type=symbol.type if isinstance(symbol, IExpression) else symbol,
                               already_generated=True)

    def visit_Raise(self, raise_node: ast.Raise) -> GeneratorData:
        """
        Visitor of the raise node

        :param raise_node: the python ast raise node
        """
        self.visit_to_map(raise_node.exc, generate=True)
        self.generator.convert_raise_exception()
        return self.build_data(raise_node)

    def visit_Try(self, try_node: ast.Try) -> GeneratorData:
        """
        Visitor of the try node

        :param try_node: the python ast try node
        """
        try_address: int = self.generator.convert_begin_try()
        try_end: Optional[int] = None
        for stmt in try_node.body:
            self.visit_to_map(stmt, generate=True)

        if len(try_node.handlers) == 1:
            handler = try_node.handlers[0]
            try_end = self.generator.convert_try_except(handler.name)
            for stmt in handler.body:
                self.visit_to_map(stmt, generate=True)

        else_address = None
        if len(try_node.orelse) > 0:
            else_start_address = self.generator.convert_begin_else(try_end)
            else_address = self.generator.bytecode_size
            for stmt in try_node.orelse:
                self.visit_to_map(stmt, generate=True)
            self.generator.convert_end_if(else_start_address)

        except_end = self.generator.convert_end_try(try_address, try_end, else_address)
        for stmt in try_node.finalbody:
            self.visit_to_map(stmt, generate=True)
        self.generator.convert_end_try_finally(except_end, try_address, len(try_node.finalbody) > 0)

        return self.build_data(try_node)

    def visit_Name(self, name_node: ast.Name) -> GeneratorData:
        """
        Visitor of a name node

        :param name_node: the python ast name identifier node
        :return: the identifier of the name
        """
        name = name_node.id
        if name not in self._symbols:
            hashed_name = self._get_unique_name(name, self._tree)
            if hashed_name not in self._symbols and hasattr(name_node, 'origin'):
                hashed_name = self._get_unique_name(name, name_node.origin)

            if hashed_name in self._symbols:
                name = hashed_name
        return self.build_data(name_node, name)

    def visit_Starred(self, starred: ast.Starred) -> GeneratorData:
        value_data = self.visit_to_generate(starred.value)
        self.generator.convert_starred_variable()

        starred_data = value_data.copy(starred)
        starred_data.already_generated = True
        return starred_data

    def visit_Attribute(self, attribute: ast.Attribute) -> GeneratorData:
        """
        Visitor of a attribute node

        :param attribute: the python ast attribute node
        :return: the identifier of the attribute
        """
        last_address = VMCodeMapping.instance().bytecode_size
        last_stack = self.generator.stack_size

        _, attr = self.generator.get_symbol(attribute.attr)
        value = attribute.value
        value_symbol = None
        value_type = None
        value_data = self.visit(value)
        need_to_visit_again = True

        if value_data.symbol_id is not None and not value_data.already_generated:
            value_id = value_data.symbol_id
            if value_data.symbol is not None:
                value_symbol = value_data.symbol
            else:
                _, value_symbol = self.generator.get_symbol(value_id)
            value_type = value_symbol.type if hasattr(value_symbol, 'type') else value_symbol

            if hasattr(value_type, 'symbols') and attribute.attr in value_type.symbols:
                attr = value_type.symbols[attribute.attr]
            elif isinstance(value_type, Package) and attribute.attr in value_type.inner_packages:
                attr = value_type.inner_packages[attribute.attr]

            if isinstance(value_symbol, UserClass):
                if isinstance(attr, Method) and attr.has_cls_or_self:
                    self.generator.convert_load_symbol(value_id)
                    need_to_visit_again = False

                if isinstance(attr, Variable):
                    self.visit_to_generate(attribute.value)
                    self.generator.convert_load_symbol(attribute.attr, class_type=value_symbol)
                    return self.build_data(attribute, symbol=attr)
        else:
            need_to_visit_again = value_data.already_generated
            self._remove_inserted_opcodes_since(last_address, last_stack)

        # the verification above only verify variables, this one will should work with literals and constants
        if isinstance(value, (ast.Constant if SYS_VERSION_INFO >= (3, 8) else (ast.Num, ast.Str, ast.Bytes))) \
                and len(attr.args) > 0 and isinstance(attr, IBuiltinMethod) and attr.has_self_argument:
            attr = attr.build(value_data.type)

        if attr is not Type.none and not hasattr(attribute, 'generate_value'):
            value_symbol_id = (value_symbol.identifier
                               if self.is_implemented_class_type(value_symbol)
                               else value_data.symbol_id)
            attribute_id = f'{value_symbol_id}{constants.ATTRIBUTE_NAME_SEPARATOR}{attribute.attr}'
            index = value_type if isinstance(value_type, Package) else None
            return self.build_data(attribute, symbol_id=attribute_id, symbol=attr, index=index)

        if isinstance(value, ast.Attribute):
            value_data = self.visit(value)
        elif hasattr(attribute, 'generate_value') and attribute.generate_value:
            current_bytecode_size = self.generator.bytecode_size
            if need_to_visit_again:
                value_data = self.visit_to_generate(attribute.value)

            result = value_data.type
            generation_result = value_data.symbol
            if result is None and value_data.symbol_id is not None:
                _, result = self.generator.get_symbol(value_data.symbol_id)
                if isinstance(result, IExpression):
                    generation_result = result
                    result = result.type

            if self.is_implemented_class_type(result):
                class_attr_id = f'{result.identifier}.{attribute.attr}'
                symbol_id = class_attr_id
                symbol = None
                result_type = None
                symbol_index = None

                if self.generator.bytecode_size > current_bytecode_size and isinstance(result, UserClass):
                    # it was generated already, don't convert again
                    generated = False
                    symbol_id = attribute.attr if isinstance(generation_result, Variable) else class_attr_id
                    result_type = result
                else:
                    index = self.generator.convert_class_symbol(result,
                                                                attribute.attr,
                                                                isinstance(attribute.ctx, ast.Load))
                    generated = True
                    symbol = result
                    if not isinstance(result, UserClass):
                        if isinstance(index, int):
                            symbol_index = index
                        else:
                            symbol_id = index

                return self.build_data(attribute,
                                       symbol_id=symbol_id, symbol=symbol,
                                       result_type=result_type, index=symbol_index,
                                       already_generated=generated)

        if value_data is not None and value_symbol is None:
            value_symbol = value_data.symbol_id

        if value_data is not None and value_data.symbol_id is not None:
            value_id = f'{value_data.symbol_id}{constants.ATTRIBUTE_NAME_SEPARATOR}{attribute.attr}'
        else:
            value_id = attribute.attr

        return self.build_data(attribute, symbol_id=value_id, symbol=value_symbol)

    def visit_Continue(self, continue_node: ast.Continue) -> GeneratorData:
        """
        :param continue_node: the python ast continue statement node
        """
        self.generator.convert_loop_continue()
        return self.build_data(continue_node)

    def visit_Break(self, break_node: ast.Break) -> GeneratorData:
        """
        :param break_node: the python ast break statement node
        """
        self.generator.convert_loop_break()
        return self.build_data(break_node)

    def visit_Constant(self, constant: ast.Constant) -> GeneratorData:
        """
        Visitor of constant values node

        :param constant: the python ast constant value node
        """
        index = self.generator.convert_literal(constant.value)
        result_type = self.get_type(constant.value)
        return self.build_data(constant, result_type=result_type, index=index, already_generated=True)

    def visit_NameConstant(self, constant: ast.NameConstant) -> GeneratorData:
        """
        Visitor of constant names node

        :param constant: the python ast name constant node
        """
        index = self.generator.convert_literal(constant.value)
        result_type = self.get_type(constant.value)
        return self.build_data(constant, result_type=result_type, index=index, already_generated=True)

    def visit_Num(self, num: ast.Num) -> GeneratorData:
        """
        Visitor of literal number node

        :param num: the python ast number node
        """
        index = self.generator.convert_literal(num.n)
        result_type = self.get_type(num.n)
        return self.build_data(num, result_type=result_type, index=index, already_generated=True)

    def visit_Str(self, string: ast.Str) -> GeneratorData:
        """
        Visitor of literal string node

        :param string: the python ast string node
        """
        index = self.generator.convert_literal(string.s)
        result_type = self.get_type(string.s)
        return self.build_data(string, result_type=result_type, index=index, already_generated=True)

    def visit_Bytes(self, bts: ast.Bytes) -> GeneratorData:
        """
        Visitor of literal bytes node

        :param bts: the python ast bytes node
        """
        index = self.generator.convert_literal(bts.s)
        result_type = self.get_type(bts.s)
        return self.build_data(bts, result_type=result_type, index=index, already_generated=True)

    def visit_Tuple(self, tup_node: ast.Tuple) -> GeneratorData:
        """
        Visitor of literal tuple node

        :param tup_node: the python ast tuple node
        """
        result_type = Type.tuple
        self._create_array(tup_node.elts, result_type)
        return self.build_data(tup_node, result_type=result_type, already_generated=True)

    def visit_List(self, list_node: ast.List) -> GeneratorData:
        """
        Visitor of literal list node

        :param list_node: the python ast list node
        """
        result_type = Type.list
        self._create_array(list_node.elts, result_type)
        return self.build_data(list_node, result_type=result_type, already_generated=True)

    def visit_Dict(self, dict_node: ast.Dict) -> GeneratorData:
        """
        Visitor of literal dict node

        :param dict_node: the python ast dict node
        """
        result_type = Type.dict
        length = min(len(dict_node.keys), len(dict_node.values))
        self.generator.convert_new_map(result_type)
        for key_value in range(length):
            self.generator.duplicate_stack_top_item()
            self.visit_to_generate(dict_node.keys[key_value])
            value_data = self.visit_to_generate(dict_node.values[key_value])
            self.generator.convert_set_item(value_data.index)

        return self.build_data(dict_node, result_type=result_type, already_generated=True)

    def visit_Pass(self, pass_node: ast.Pass) -> GeneratorData:
        """
        Visitor of pass node

        :param pass_node: the python ast dict node
        """
        # only generates if the scope is a function
        result_type = None
        generated = False

        if isinstance(self.current_method, Method):
            self.generator.insert_nop()
            generated = True

        return self.build_data(pass_node, result_type=result_type, already_generated=generated)

    def _create_array(self, values: List[ast.AST], array_type: IType):
        """
        Creates a new array from a literal sequence

        :param values: list of values of the new array items
        """
        length = len(values)
        if length > 0:
            for value in reversed(values):
                self.visit_to_generate(value)
        self.generator.convert_new_array(length, array_type)



from typing import List

from boa3 import constants
from boa3_test.tests.test_classes.testcontract import TestContract


class ContractCollection:
    def __init__(self):
        self._internal_list: List[TestContract] = []

    def append(self, new_contract: TestContract):
        return self._internal_list.append(new_contract)

    def remove(self, contract: TestContract):
        return self._internal_list.remove(contract)

    def pop(self, index: int):
        return self._internal_list.pop(index)

    def clear(self):
        return self._internal_list.clear()

    def copy(self):
        collection_copy = ContractCollection()
        collection_copy._internal_list = self._internal_list.copy()
        return collection_copy

    def __len__(self):
        return len(self._internal_list)

    def __contains__(self, item) -> bool:
        return any(contract.is_valid_identifier(item) for contract in self._internal_list)

    def __getitem__(self, item) -> TestContract:
        from boa3.neo3.core.types import UInt160
        if isinstance(item, UInt160):
            item = str(item)
        elif isinstance(item, bytes) and len(item) == constants.SIZE_OF_INT160:
            item = str(UInt160(item))
        return next((contract for contract in self._internal_list if contract.is_valid_identifier(item)), None)

    def __iter__(self):
        return self._internal_list.__iter__()

    def __str__(self):
        return str(self._internal_list)

    def __repr__(self):
        return self._internal_list.__repr__()



from typing import Dict, Optional

from boa3.model.builtin.interop.nativecontract import OracleMethod
from boa3.model.variable import Variable


class OracleGetPriceMethod(OracleMethod):

    def __init__(self):
        from boa3.model.type.type import Type

        identifier = 'get_price'
        syscall = 'getPrice'
        args: Dict[str, Variable] = {}

        super().__init__(identifier, syscall, args, return_type=Type.int)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None



from __future__ import annotations

from typing import Any, Dict

from boa3.neo import from_hex_str
from boa3.neo3.core.types import UInt160
from boa3_test.tests.test_classes.witnessscope import WitnessScope


class Signer:
    def __init__(self, account: UInt160, scopes: WitnessScope = WitnessScope.CalledByEntry):
        self._account: UInt160 = account
        self._scopes: WitnessScope = scopes

    @property
    def account(self) -> UInt160:
        return self._account

    @property
    def scopes(self) -> WitnessScope:
        return self._scopes

    def to_json(self) -> Dict[str, Any]:
        return {
            'account': str(self._account),
            'scopes': self._scopes.neo_name()
        }

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Signer:
        account_hex = json['account']
        account = UInt160(from_hex_str(account_hex))
        scopes = WitnessScope.get_from_neo_name(json['scopes']) if 'scopes' in json else WitnessScope.CalledByEntry
        return cls(account, scopes)

    def __str__(self) -> str:
        return self._account.__str__()

    def __eq__(self, other) -> bool:
        return isinstance(other, Signer) and self._account == other._account



from typing import List, Tuple

from boa3.model.builtin.method.printmethod import PrintMethod
from boa3.neo.vm.opcode.Opcode import Opcode


class PrintIntMethod(PrintMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        arg_value = Type.int

        super().__init__(arg_value)

    @property
    def print_value_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        if self._print_value_opcodes is None:
            from boa3.model.builtin.interop.interop import Interop
            itoa_method = Interop.Itoa
            self._print_value_opcodes = (
                [
                    Opcode.get_push_and_data(1),
                    (Opcode.PACK, b'')
                ] +
                itoa_method.opcode.copy()
            )

        return super().print_value_opcodes



from boa3.model.builtin.decorator.builtindecorator import IBuiltinDecorator


class InstanceMethodDecorator(IBuiltinDecorator):
    def __init__(self):
        identifier = '-instancemethod'  # MUST NOT be used from smart contracts
        super().__init__(identifier)

    @property
    def has_cls_or_self(self) -> bool:
        return True



from typing import Dict

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class CreateStandardAccountMethod(InteropMethod):

    def __init__(self):
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        from boa3.model.type.collection.sequence.ecpointtype import ECPointType

        identifier = 'create_standard_account'
        syscall = 'System.Contract.CreateStandardAccount'
        args: Dict[str, Variable] = {
            'pub_key': Variable(ECPointType.build())
        }
        super().__init__(identifier, syscall, args, return_type=UInt160Type.build())



from typing import Dict, List, Tuple

from boa3.model.builtin.interop.nativecontract import NeoContractMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class UnVoteMethod(NeoContractMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'un_vote'
        native_identifier = 'vote'  # un_vote is not neo native
        args: Dict[str, Variable] = {
            'account': Variable(UInt160Type.build()),
        }

        super().__init__(identifier, native_identifier, args, return_type=Type.bool)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return (
            [
                (Opcode.UNPACK, b''),
                (Opcode.DROP, b''),
                (Opcode.PUSHNULL, b''),
                (Opcode.SWAP, b''),
                (Opcode.PUSH2, b''),
                (Opcode.PACK, b''),
            ] +
            super().opcode
        )



from __future__ import annotations

from typing import Optional

from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.opcode.OpcodeInformation import OpcodeInformation
from boa3.neo.vm.type.Integer import Integer


class VMCode:
    """
    Represents a Neo VM code

    :ivar info: the opcode information of the code
    :ivar data: the data in bytes of the code. Empty byte array by default.
    """

    def __init__(self, op_info: OpcodeInformation, data: bytes = None):
        """
        :param op_info: information of the opcode of the code
        :param data: the data in bytes of the code. Empty byte array by default.
        """
        if data is None:
            data = bytes(op_info.data_len)
        self._info: OpcodeInformation = op_info
        self._target: Optional[VMCode] = None
        self._data: bytes = data

    @property
    def info(self) -> OpcodeInformation:
        """
        Gets the Neo VM opcode information

        :return: the opcode of the code.
        """
        return self._info

    @property
    def data(self) -> bytes:
        """
        Gets the Neo VM data of the code

        :return: the formatted data in bytes of the code.
        """
        data: bytearray = bytearray(self.raw_data)
        info = self.info

        if len(data) < info.data_len:
            data.extend(bytes(info.data_len))
        elif len(data) > info.max_data_len:
            data = data[:info.max_data_len]
        return data

    @property
    def raw_data(self) -> bytes:
        """
        Gets the Neo VM raw data of the code

        :return: the unformatted data in bytes of the code.
        """
        if self.target is None:
            return self._data
        else:
            from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping
            code_mapping = VMCodeMapping.instance()
            self_start = code_mapping.get_start_address(self)
            target_start = code_mapping.get_start_address(self.target)

            if self_start == target_start:
                return self._data

            return (Integer(target_start - self_start)
                    .to_byte_array(signed=True, min_length=self._info.data_len))

    @property
    def size(self) -> int:
        return len(self._info.opcode) + len(self.data)

    @property
    def opcode(self) -> Opcode:
        """
        Gets the Neo VM opcode of the code

        :return: the opcode of the code
        """
        return self.info.opcode

    @property
    def target(self) -> VMCode:
        """
        Gets the target code of this code

        :return: the target code if this is a control code. None otherwise
        :rtype: VMCode
        """
        return self._target if self.opcode.has_target() else None

    def set_target(self, target_code):
        """
        Set the target code if this instruction's opcode requires a target

        :param target_code: the target code of this instruction
        :type target_code: VMCode
        """
        if self.opcode.has_target():
            self._target = target_code

    def __str__(self) -> str:
        return self.opcode.name + ' ' + self.data.hex()

    def __repr__(self) -> str:
        return str(self)



from __future__ import annotations

import hashlib
from typing import Tuple

from boa3.neo3.core import Size as s, serialization, types, utils


class Version(serialization.ISerializable):
    """
    Represents the version number of an assembly
    """

    def __init__(self, major: int = 0, minor: int = 0, build: int = 0, revision: int = 0):
        """
        Args:
            major: non interchangeable assembly.
            minor: significant enhancements with backwards compatibility.
            build: recompilation of the same source with possible other compiler or on other platform.
            revision: fully interchangeable. Can be used for example for security fixes.
        """
        if major < 0 or minor < 0 or build < 0 or revision < 0:
            raise ValueError("Negative version numbers are not allowed")

        if major > 255 or minor > 255 or build > 255 or revision > 255:
            raise ValueError("Version numbers cannot exceed 255")

        self.major = major
        self.minor = minor
        self.build = build
        self.revision = revision

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.major == other.major
                and self.minor == other.minor
                and self.build == other.build
                and self.revision == other.revision)

    def __len__(self):
        return s.uint64 + s.uint64 + s.uint64 + s.uint64

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        version_str = "{0}.{1}.{2}.{3}".format(self.major,
                                               self.minor,
                                               self.build,
                                               self.revision)
        from boa3.neo.vm.type.String import String
        version_bytes = String(version_str).to_bytes() + bytes(s.uint64 * 4 - len(version_str))
        writer.write_bytes(version_bytes)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        version_str = reader.read_bytes(s.uint64 * 4).decode('utf-8')

        import re
        version_str = re.sub(r'\x00+', '', version_str)

        major, minor, build, revision = version_str.split('.')
        self.major = int(major)
        self.minor = int(minor)
        self.build = int(build)
        self.revision = int(revision)

    @classmethod
    def _parse_component(self, c: str) -> Tuple[bool, int]:
        try:
            r = int(c)
        except ValueError:
            return False, -1
        if r < 0:
            return False, -1
        if r > 255:
            return False, -1
        return True, r

    @classmethod
    def from_string(cls, input: str) -> Version:
        """
        Parse an instance out of a string.

        Args:
            input: string representing a version number following the format `Major.Minor[.build[.revision]]`.
            Each version part must fit in the range >= 0 <= 255.

        Raises:
            ValueError: if the input cannot be successfully parsed.
        """
        parts = input.split('.')
        if len(parts) < 2 or len(parts) > 4:
            raise ValueError(f"Cannot parse version from: {input}")

        success, major = Version._parse_component(parts[0])
        if not success:
            raise ValueError(f"Cannot parse major field from: {parts[0]}")
        success, minor = Version._parse_component(parts[1])
        if not success:
            raise ValueError(f"Cannot parse minor field from: {parts[1]}")

        if len(parts) > 2:
            success, build = Version._parse_component(parts[2])
            if not success:
                raise ValueError(f"Cannot parse build field from: {parts[2]}")
        else:
            build = 0

        if len(parts) > 3:
            success, revision = Version._parse_component(parts[3])
            if not success:
                raise ValueError(f"Cannot parse revision field from: {parts[3]}")
        else:
            revision = 0
        return cls(major, minor, build, revision)


class NEF(serialization.ISerializable):
    def __init__(self, compiler_name: str = None, version: Version = None, script: bytes = None):
        self.magic = 0x3346454E
        if compiler_name is None:
            self.compiler = 'unknown'
        else:
            self.compiler = compiler_name[:32]
        self.version = version if version else Version()
        self.script = script if script else b''
        self.checksum = self.compute_checksum()

    def __len__(self):
        return (
            s.uint32  # magic
            + 32  # compiler
            + (s.uint64 * 4)  # version
            + 2  # reserve
            + utils.get_var_size(bytes())  # TODO: method tokens
            + 2  # reserve
            + utils.get_var_size(self.script)
            + s.uint32)  # checksum

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.magic == other.magic
                and self.compiler == other.compiler
                and self.version == other.version
                and self.script == other.script
                and self.checksum == other.checksum)

    @property
    def compiler_with_version(self) -> bytes:
        result = '{0}-'.format(self.compiler).encode('utf-8') + self.version.to_array()
        return result[:64] + bytes(64 - len(result))

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_uint32(self.magic)
        writer.write_bytes(self.compiler_with_version)

        writer.write_uint16(0)    # 2 reserved bytes
        writer.write_var_bytes(bytes())   # TODO: method tokens
        writer.write_uint16(0)    # 2 reserved bytes

        writer.write_var_bytes(self.script)
        writer.write_bytes(self.checksum)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.magic = reader.read_uint32()
        compiler_with_version = reader.read_bytes(64).decode('utf-8')
        self.compiler, version = compiler_with_version.rsplit('-', maxsplit=1)
        self.version = Version.deserialize_from_bytes(version[:32].encode('utf-8'))

        assert reader.read_uint16() == 0    # 2 reserved bytes
        reader.read_var_int(128)            # TODO: method tokens
        assert reader.read_uint16() == 0    # 2 reserved bytes

        self.script = reader.read_var_bytes()
        self.checksum = reader.read_bytes(4)
        if self.checksum != self.compute_checksum():
            raise ValueError("Deserialization error - invalid checksum")

    def script_to_array(self):
        from boa3.neo3.core.serialization import BinaryWriter
        with BinaryWriter() as bw:
            bw.write_var_bytes(self.script)
            return bw._stream.getvalue()

    def tokens_to_array(self):
        from boa3.neo3.core.serialization import BinaryWriter
        with BinaryWriter() as bw:
            bw.write_var_bytes(bytes())
            return bw._stream.getvalue()

    def compute_checksum(self) -> bytes:
        data = (self.magic.to_bytes(4, 'little')
                + self.compiler_with_version
                + bytes(2)  # reserved bytes
                + self.tokens_to_array()   # TODO: method tokens
                + bytes(2)  # reserved bytes
                + self.script_to_array())

        return hashlib.sha256(hashlib.sha256(data).digest()).digest()[:s.uint32]

    def compute_script_hash(self) -> types.UInt160:
        hash = hashlib.new('ripemd160', hashlib.sha256(self.script).digest()).digest()
        return types.UInt160(data=hash)



import ast
from typing import Dict, List

from boa3.model import set_internal_call
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class CallMethod(InteropMethod):

    def __init__(self):
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        from boa3.model.builtin.interop.contract.callflagstype import CallFlagsType
        from boa3.model.type.type import Type
        identifier = 'call_contract'
        syscall = 'System.Contract.Call'

        call_flags = CallFlagsType.build()
        args: Dict[str, Variable] = {
            'script_hash': Variable(UInt160Type.build()),
            'method': Variable(Type.str),
            'args': Variable(Type.sequence),  # TODO: change when *args is implemented
            'call_flags': Variable(call_flags)
        }
        args_default = ast.parse("{0}".format(Type.sequence.default_value)
                                 ).body[0].value
        call_flags_default = set_internal_call(ast.parse("{0}.{1}".format(call_flags.identifier,
                                                                          call_flags.default_value.name)
                                                         ).body[0].value)

        super().__init__(identifier, syscall, args, defaults=[args_default, call_flags_default], return_type=Type.any)

    @property
    def generation_order(self) -> List[int]:
        """
        Gets the indexes order that need to be used during code generation.
        If the order for generation is the same as inputted in code, returns reversed(range(0,len_args))

        :return: Index order for code generation
        """
        indexes = super().generation_order
        args_index = list(self.args).index('args')

        if indexes[0] != args_index:
            # context must be the last generated argument
            indexes.remove(args_index)
            indexes.insert(0, args_index)

        return indexes



from typing import Any, Dict, List, Optional, Sized, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.annotation.metatype import MetaType, metaType
from boa3.model.type.anytype import anyType
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class CastTypeMethod(IBuiltinMethod):

    def __init__(self, cast_to_type: IType = None, origin_type: IType = None):
        # if the type to be casted is not value, there's an error in the signature
        if cast_to_type is None:
            type_value = metaType
        elif isinstance(cast_to_type, MetaType):
            type_value = cast_to_type if cast_to_type.has_meta_type else metaType
        else:
            type_value = MetaType.build(cast_to_type)

        if origin_type is None:
            origin_type = anyType
        elif isinstance(origin_type, MetaType):
            origin_type = origin_type.meta_type if origin_type.has_meta_type else anyType

        identifier = 'cast'
        args: Dict[str, Variable] = {'typ': Variable(type_value),
                                     'val': Variable(anyType)}
        super().__init__(identifier, args,
                         return_type=cast_to_type.meta_type if isinstance(cast_to_type, MetaType) else anyType)
        self._origin_type: IType = origin_type

    @property
    def identifier(self) -> str:
        if self.is_supported:
            identifier = (self.typ_arg.type.meta_id
                          if hasattr(self.typ_arg.type, 'meta_id')
                          else self.typ_arg.type.identifier)
            return '-{0}_{1}'.format(self._identifier, identifier)
        else:
            return self._identifier

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    @property
    def is_supported(self) -> bool:
        return (isinstance(self.typ_arg, Variable)
                and isinstance(self.typ_arg.type, MetaType)
                and self.typ_arg.type.has_meta_type)

    def not_supported_str(self, callable_id: str) -> str:
        return '{0}({1})'.format(callable_id,
                                 ','.join([arg.type.identifier if isinstance(arg, Variable) else 'unknown'
                                           for arg in self.args.values()]))

    @property
    def is_cast(self) -> bool:
        return True

    @property
    def cast_types(self) -> Optional[Tuple[IType, IType]]:
        origin_type = self._origin_type
        target_type = (self.typ_arg.type.meta_type
                       if hasattr(self.typ_arg.type, 'meta_type')
                       else self.typ_arg.type)
        return origin_type, target_type

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return []

    @property
    def _args_on_stack(self) -> int:
        return 1  # the implementation is the same as x = arg

    @property
    def args_on_stack(self) -> int:
        return 1  # the implementation is the same as x = arg

    @property
    def generation_order(self) -> List[int]:
        # type should not be converted
        indexes = super().generation_order
        typ_index = list(self.args).index('typ')

        if typ_index in indexes:
            indexes.remove(typ_index)

        return indexes

    @property
    def _body(self) -> Optional[str]:
        return None

    @property
    def typ_arg(self) -> Variable:
        return self.args['typ'] if isinstance(self.args['typ'], Variable) else Variable(anyType)

    @property
    def val_arg(self) -> Variable:
        return self.args['val']

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, Sized) and len(value) == len(self.args):
            cast_to = value[0]
            cast_from = value[1]

            if isinstance(cast_to, IType):
                if isinstance(cast_from, IType):
                    return CastTypeMethod(cast_to, cast_from)
                else:
                    return CastTypeMethod(cast_to)
        return super().build(value)



from typing import Any, Set

from boa3.model.type.collection.icollection import ICollectionType
from boa3.model.type.itype import IType


class GenericCollectionType(ICollectionType):
    """
    An class used to represent a generic Python collection type
    """

    def __init__(self, keys_type: Set[IType] = None, values_type: Set[IType] = None):
        identifier: str = 'collection'
        keys_type = self.filter_types(keys_type)
        values_type = self.filter_types(values_type)
        super().__init__(identifier, keys_type, values_type)

    def is_valid_key(self, key_type: IType) -> bool:
        return key_type == self.valid_key

    @property
    def valid_key(self) -> IType:
        return self.key_type

    @property
    def is_generic(self) -> bool:
        return True

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, ICollectionType)

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            if isinstance(value, dict):
                keys = list(value.keys())
                values = list(value.values())
            else:
                keys = value.key_type
                values = value.value_type

            keys_types: Set[IType] = cls.get_types(keys)
            values_types: Set[IType] = cls.get_types(values)
            return cls(keys_types, values_types)

    def __hash__(self):
        return hash(self.identifier)



__all__ = ['GetAccountStateMethod',
           'GetAllCandidatesMethod',
           'GetCandidatesMethod',
           'GetCandidateVoteMethod',
           'GetCommitteeMethod',
           'GetGasPerBlockMethod',
           'GetNextBlockValidatorsMethod',
           'RegisterCandidateMethod',
           'UnclaimedGasMethod',
           'UnregisterCandidateMethod',
           'UnVoteMethod',
           'VoteMethod',
           ]

from boa3.model.builtin.native.neo_contract_methods.getaccountstatemethod import GetAccountStateMethod
from boa3.model.builtin.native.neo_contract_methods.getallcandidatesmethod import GetAllCandidatesMethod
from boa3.model.builtin.native.neo_contract_methods.getcandidatesmethod import GetCandidatesMethod
from boa3.model.builtin.native.neo_contract_methods.getcandidatevotemethod import GetCandidateVoteMethod
from boa3.model.builtin.native.neo_contract_methods.getcommitteemethod import GetCommitteeMethod
from boa3.model.builtin.native.neo_contract_methods.getgasperblockmethod import GetGasPerBlockMethod
from boa3.model.builtin.native.neo_contract_methods.getnextblockvalidators import GetNextBlockValidatorsMethod
from boa3.model.builtin.native.neo_contract_methods.registercandidatemethod import RegisterCandidateMethod
from boa3.model.builtin.native.neo_contract_methods.unclaimedgasmethod import UnclaimedGasMethod
from boa3.model.builtin.native.neo_contract_methods.unregistercandidatemethod import UnregisterCandidateMethod
from boa3.model.builtin.native.neo_contract_methods.unvote import UnVoteMethod
from boa3.model.builtin.native.neo_contract_methods.vote import VoteMethod



from typing import Dict

from boa3.model.builtin.math import *
from boa3.model.callable import Callable
from boa3.model.symbol import ISymbol


class Math:

    # python math methods
    Pow = PowMethod()
    Sqrt = SqrtMethod()

    @classmethod
    def all_functions(cls) -> Dict[str, Callable]:
        from boa3.model.builtin.builtincallable import IBuiltinCallable

        functions = [
            cls.Pow,
            cls.Sqrt,
        ]

        return {method._identifier: method for method in functions if isinstance(method, IBuiltinCallable)}

    @classmethod
    def get_methods_from_math_lib(cls) -> Dict[str, ISymbol]:
        import math
        from inspect import getmembers, isbuiltin

        method_symbols: Dict[str, ISymbol] = {}
        all_functions = getmembers(math, isbuiltin)

        for m_id, method in all_functions:
            method_id: str = m_id if m_id in cls.all_functions() else m_id.lower()
            if method_id in cls.all_functions():
                method_symbols[m_id] = cls.all_functions()[method_id]

        return method_symbols



from __future__ import annotations

import ast
from abc import ABC
from typing import Dict, List, Optional, Set, Tuple

from boa3.model import set_internal_call
from boa3.model.expression import IExpression
from boa3.model.type.type import IType, Type
from boa3.model.variable import Variable
from boa3.neo.vm.VMCode import VMCode


class Callable(IExpression, ABC):
    """
    A class used to represent a function or a class method

    :ivar args: a dictionary that maps each arg with its name. Empty by default.
    :ivar is_public: a boolean value that specifies if the method is public. False by default.
    :ivar return_type: the return type of the method. None by default.
    """

    def __init__(self, args: Dict[str, Variable] = None,
                 vararg: Optional[Tuple[str, Variable]] = None,
                 kwargs: Optional[Dict[str, Variable]] = None,
                 defaults: List[ast.AST] = None,
                 return_type: IType = Type.none, is_public: bool = False,
                 decorators: List[Callable] = None,
                 external_name: str = None,
                 is_safe: bool = False,
                 origin_node: Optional[ast.AST] = None):

        if args is None:
            args = {}
        self.args: Dict[str, Variable] = args.copy()

        if not isinstance(defaults, list):
            defaults = []
        self.defaults: List[ast.AST] = defaults

        self._vararg: Optional[Tuple[str, Variable]] = None
        if (isinstance(vararg, tuple) and len(vararg) == 2
                and isinstance(vararg[0], str) and isinstance(vararg[1], Variable)):

            from boa3.model.type.typeutils import TypeUtils

            vararg_id, vararg_var = vararg
            if vararg_var.type is not Type.any:
                default_code = "{0}({1}, {2})".format(TypeUtils.cast.raw_identifier,
                                                      Type.tuple.build_collection(vararg_var.type),
                                                      Type.tuple.default_value)
            else:
                default_code = "{0}".format(Type.tuple.default_value)

            default_value = set_internal_call(ast.parse(default_code).body[0].value)

            self.args[vararg_id] = Variable(Type.tuple.build_collection([vararg_var.type]))
            self.defaults.append(default_value)
            self._vararg = vararg

        if kwargs is None:
            kwargs = {}
        self._kwargs: Dict[str, Variable] = kwargs.copy()

        self.return_type: IType = return_type

        if decorators is None:
            decorators = []
        from boa3.model.decorator import IDecorator
        self.decorators: List[IDecorator] = [decorator for decorator in decorators
                                             if isinstance(decorator, IDecorator)]

        from boa3.model.builtin.decorator import PublicDecorator
        public_decorator = next((decorator for decorator in self.decorators
                                 if isinstance(decorator, PublicDecorator)),
                                None)

        self.is_public: bool = is_public or public_decorator is not None
        if self.is_public:
            if isinstance(public_decorator, PublicDecorator):
                external_name = public_decorator.name
            elif self.defined_by_entry:
                external_name = None

        self.external_name: Optional[str] = external_name
        self.is_safe: bool = is_safe or (isinstance(public_decorator, PublicDecorator) and public_decorator.safe)

        self._self_calls: Set[ast.AST] = set()

        super().__init__(origin_node)

        self.init_address: Optional[int] = None
        self.init_bytecode: Optional[VMCode] = None
        self.init_defaults_bytecode: Optional[VMCode] = None
        self.end_bytecode: Optional[VMCode] = None

    @property
    def type(self) -> IType:
        return self.return_type

    @property
    def symbols(self) -> Dict[str, Variable]:
        """
        Gets all the symbols in the method

        :return: a dictionary that maps each symbol in the module with its name
        """
        return self.args.copy()

    @property
    def args_without_default(self) -> Dict[str, Variable]:
        num_defaults = len(self.defaults)
        if num_defaults > 0:
            return {key: self.args[key] for key in list(self.args.keys())[:-num_defaults]}
        return self.args

    @property
    def has_cls_or_self(self) -> bool:
        return any(decorator.has_cls_or_self for decorator in self.decorators)

    @property
    def cls_or_self_type(self) -> Optional[IType]:
        if not self.has_cls_or_self or len(self.args) == 0:
            return None

        return list(self.args.values())[0].type

    @property
    def has_starred_argument(self) -> bool:
        return self._vararg is not None

    @property
    def start_address(self) -> Optional[int]:
        """
        Gets the address where this method starts in the bytecode

        :return: the first address of the method
        """
        if self.init_bytecode is None and self.init_defaults_bytecode is None:
            return self.init_address
        else:
            from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping
            return VMCodeMapping.instance().get_start_address(self.init_bytecode)

    @property
    def start_bytecode(self) -> Optional[VMCode]:
        return (self.init_defaults_bytecode if len(self.defaults) > 0
                else self.init_bytecode)

    @property
    def end_address(self) -> Optional[int]:
        """
        Gets the address of this method's last operation in the bytecode

        :return: the last address of the method
        """
        if self.end_bytecode is None:
            return self.start_address
        else:
            from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping
            return VMCodeMapping.instance().get_end_address(self.end_bytecode)

    @property
    def is_called(self) -> bool:
        return len(self._self_calls) > 0

    def reset_calls(self):
        self._self_calls.clear()

    @property
    def is_compiled(self) -> bool:
        return self.start_address is not None and self.end_address is not None

    def add_call_origin(self, origin: ast.AST) -> bool:
        try:
            self._self_calls.add(origin)
            return True
        except BaseException:
            return False

    def __str__(self) -> str:
        args_types: List[str] = [str(arg.type) for arg in self.args.values()]
        if self.return_type is not Type.none:
            signature = '({0}) -> {1}'.format(', '.join(args_types), self.return_type)
        else:
            signature = '({0})'.format(', '.join(args_types))
        public = 'public ' if self.is_public else ''
        return '{0}{1}'.format(public, signature)

    def __repr__(self) -> str:
        name = self.identifier if hasattr(self, 'identifier') else self.__class__.__name__
        return f'{name}{str(self)}'



import ast
from typing import Dict, List, Tuple

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.builtin.method import IBuiltinMethod
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class NativeContractMethod(InteropMethod):

    def __init__(self, native_contract_script_hash_method: IBuiltinMethod, identifier: str, syscall: str,
                 args: Dict[str, Variable] = None, defaults: List[ast.AST] = None, return_type: IType = None):
        super().__init__(identifier, syscall, args, defaults, return_type)
        self.native_contract_script_hash_method = native_contract_script_hash_method

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.builtin.interop.interop import Interop
        from boa3.model.type.type import Type
        from boa3.neo.vm.type.Integer import Integer
        from boa3.neo.vm.type.String import String
        from boa3.neo3.contracts import CallFlags

        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        flags_opcode = [
            (Opcode.PUSHDATA1, Integer(len(call_flags)).to_byte_array() + call_flags)
        ]
        method = String(self._sys_call).to_bytes()
        method_opcode = [
            (Opcode.PUSHDATA1, Integer(len(method)).to_byte_array(min_length=1) + method)
        ]
        drop_if_void_opcode = [
            (Opcode.DROP, b'')
        ] if self.return_type is Type.none else []

        return (flags_opcode
                + method_opcode
                + self.native_contract_script_hash_method.opcode
                + Interop.CallContract.opcode
                + drop_if_void_opcode
                )

    @property
    def pack_arguments(self) -> bool:
        return True

    @property
    def method_name(self) -> str:
        return self._sys_call



from typing import Any

from boa3.builtin.type import UInt160


class GAS:
    """
    A class used to represent the GAS native contract
    """

    @classmethod
    def symbol(cls) -> str:
        """
        Gets the symbol of GAS.

        :return: the GAS string.
        :rtype: str
        """
        pass

    @classmethod
    def decimals(cls) -> int:
        """
        Gets the amount of decimals used by GAS.

        :return: the number 8.
        :rtype: int
        """
        pass

    @classmethod
    def totalSupply(cls) -> int:
        """
        Gets the total token supply deployed in the system.

        :return: the total token supply deployed in the system.
        :rtype: int
        """
        pass

    @classmethod
    def balanceOf(cls, account: UInt160) -> int:
        """
        Get the current balance of an address.

        :param account: the account's address to retrieve the balance for
        :type account: UInt160
        :return: the account's balance
        :rtype: int
        """
        pass

    @classmethod
    def transfer(cls, from_address: UInt160, to_address: UInt160, amount: int, data: Any = None) -> bool:
        """
        Transfers an amount of GAS from one account to another

        If the method succeeds, it will fire the `Transfer` event and must return true, even if the amount is 0,
        or from and to are the same address.

        :param from_address: the address to transfer from
        :type from_address: UInt160
        :param to_address: the address to transfer to
        :type to_address: UInt160
        :param amount: the amount of GAS to transfer
        :type amount: int
        :param data: whatever data is pertinent to the onNEP17Payment method
        :type data: Any

        :return: whether the transfer was successful
        :rtype: bool
        :raise Exception: raised if `from_address` or `to_address` length is not 20 or if `amount` is less than zero.
        """
        pass



import ast
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

from boa3 import constants
from boa3.analyser.astanalyser import IAstAnalyser
from boa3.analyser.builtinfunctioncallanalyser import BuiltinFunctionCallAnalyser
from boa3.analyser.model.optimizer import Undefined, UndefinedType
from boa3.analyser.model.symbolscope import SymbolScope
from boa3.exception import CompilerError, CompilerWarning
from boa3.model.attribute import Attribute
from boa3.model.builtin.builtin import Builtin
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.callable import Callable
from boa3.model.expression import IExpression
from boa3.model.imports.importsymbol import Import
from boa3.model.imports.package import Package
from boa3.model.method import Method
from boa3.model.module import Module
from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.binaryop import BinaryOp
from boa3.model.operation.operation import IOperation
from boa3.model.operation.operator import Operator
from boa3.model.operation.unary.unaryoperation import UnaryOperation
from boa3.model.operation.unaryop import UnaryOp
from boa3.model.property import Property
from boa3.model.symbol import ISymbol
from boa3.model.type.annotation.metatype import MetaType
from boa3.model.type.classes.classtype import ClassType
from boa3.model.type.classes.pythonclass import PythonClass
from boa3.model.type.classes.userclass import UserClass
from boa3.model.type.collection.icollection import ICollectionType as Collection
from boa3.model.type.type import IType, Type
from boa3.model.type.typeutils import TypeUtils
from boa3.model.variable import Variable


class TypeAnalyser(IAstAnalyser, ast.NodeVisitor):
    """
    This class is responsible for the type checking of the code

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.

    :ivar type_errors: a list with the found type errors. Empty by default.
    :ivar modules: a list with the analysed modules. Empty by default.
    :ivar symbols: a dictionary that maps the global symbols.
    """

    def __init__(self, analyser, symbol_table: Dict[str, ISymbol], log: bool = False):
        super().__init__(analyser.ast_tree, analyser.filename, log=log)
        self.type_errors: List[Exception] = []
        self.modules: Dict[str, Module] = {}
        self.symbols: Dict[str, ISymbol] = symbol_table

        self._current_class: UserClass = None
        self._current_method: Method = None
        self._scope_stack: List[SymbolScope] = []

        self._super_calls: List[IBuiltinMethod] = []
        self.visit(self._tree)

    @property
    def _current_method_id(self) -> str:
        """
        Get the string identifier of the current method

        :return: The name identifier of the method. If the current method is None, returns None.
        :rtype: str or None
        """
        if self._current_method in self.symbols.values():
            index = list(self.symbols.values()).index(self._current_method)
            return list(self.symbols.keys())[index]

    @property
    def _current_scope(self) -> Optional[SymbolScope]:
        return self._scope_stack[-1] if len(self._scope_stack) > 0 else None

    @property
    def _modules_symbols(self) -> Dict[str, ISymbol]:
        """
        Gets all the symbols in the modules scopes.

        :return: Returns a dictionary that maps the modules symbols.
        """
        symbols = {}
        for module in self.modules.values():
            symbols.update(module.symbols)
        return symbols

    def get_symbol(self, symbol_id: str,
                   is_internal: bool = False,
                   check_raw_id: bool = False,
                   origin_node: ast.AST = None) -> Optional[ISymbol]:
        if symbol_id is None:
            return None
        if isinstance(symbol_id, ISymbol) and not isinstance(symbol_id, (IType, IExpression)):
            return symbol_id
        if not isinstance(symbol_id, str):
            return Variable(self.get_type(symbol_id))

        if len(self._scope_stack) > 0:
            for scope in self._scope_stack[::-1]:
                if symbol_id in scope:
                    return scope[symbol_id]

                if check_raw_id:
                    found_symbol = self._search_by_raw_id(symbol_id, list(scope.symbols.values()))
                    if found_symbol is not None:
                        return found_symbol

        if self._current_method is not None and symbol_id in self._current_method.symbols:
            # the symbol exists in the local scope
            return self._current_method.symbols[symbol_id]
        elif self._current_class is not None and symbol_id in self._current_class.symbols:
            # the symbol exists in the class
            return self._current_class.symbols[symbol_id]
        elif symbol_id in self.modules:
            # the symbol exists in the modules scope
            return self.modules[symbol_id]

        if self._current_method is not None:
            cur_symbols = self._current_method.symbols
        elif self._current_class is not None:
            cur_symbols = self._current_class.symbols
        else:
            cur_symbols = self.modules

        if check_raw_id:
            found_symbol = self._search_by_raw_id(symbol_id, list(cur_symbols.values()))
            if found_symbol is not None:
                return found_symbol

        return super().get_symbol(symbol_id, is_internal, check_raw_id, origin_node)

    def new_local_scope(self, symbols: Dict[str, ISymbol] = None):
        if symbols is None:
            symbols = self._current_scope.symbols if self._current_scope is not None else {}

        self._scope_stack.append(SymbolScope(symbols))

    def pop_local_scope(self) -> SymbolScope:
        return self._scope_stack.pop()

    def visit_Module(self, module: ast.Module):
        """
        Visitor of the module node

        Performs the type checking in the body of the method

        :param module: the python ast module node
        """
        for stmt in module.body:
            self.visit(stmt)

    def visit_ClassDef(self, node: ast.ClassDef) -> Any:
        if node.name in self.symbols:
            class_symbol = self.symbols[node.name]
            if isinstance(class_symbol, UserClass):
                self._current_class = class_symbol

        self.generic_visit(node)
        self._current_class = None

    def visit_FunctionDef(self, function: ast.FunctionDef):
        """
        Visitor of the function node

        Performs the type checking in the body of the function

        :param function: the python ast function definition node
        """
        self.visit(function.args)
        symbols = self.symbols if self._current_class is None else self._current_class.symbols
        method = symbols[function.name]

        from boa3.model.event import Event
        if isinstance(method, Property):
            method = method.getter

        if isinstance(method, Method):
            self._current_method = method
            self.new_local_scope({var_id: var for var_id, var in method.symbols.items()
                                  if isinstance(var, Variable) and var.type is not UndefinedType})

            if self._current_class is not None and self._current_class.is_interface and len(function.body) > 0:
                first_instruction = function.body[0]
                if not isinstance(first_instruction, ast.Pass):
                    self._log_warning(CompilerWarning.UnreachableCode(first_instruction.lineno,
                                                                      first_instruction.col_offset))
            else:
                for stmt in function.body:
                    self.visit(stmt)

            if method.is_init:
                self._check_base_init_call(function)
            else:
                self._validate_return(function)

            method_scope = self.pop_local_scope()
            for symbol_id, symbol in self._current_method.symbols.items():
                if isinstance(symbol, Variable) and symbol.type is UndefinedType and symbol_id in method_scope:
                    new_scope_symbol = method_scope[symbol_id]
                    if hasattr(new_scope_symbol, 'type') and new_scope_symbol.type is not UndefinedType:
                        symbol.set_type(new_scope_symbol.type)

            self._current_method = None
        elif (isinstance(method, Event)  # events don't have return
              and function.returns is not None):
            return_type = self.get_type(function.returns)
            if return_type is not Type.none:
                self._log_error(
                    CompilerError.MismatchedTypes(line=function.lineno, col=function.col_offset,
                                                  expected_type_id=Type.none.identifier,
                                                  actual_type_id=return_type.identifier)
                )

    def visit_arguments(self, arguments: ast.arguments):
        """
        Verifies if each argument of a function has a type annotation

        :param arguments: the python ast function arguments node
        """
        for arg in arguments.args:
            self.visit(arg)

        # continue to walk through the tree
        self.generic_visit(arguments)

    def visit_arg(self, arg: ast.arg):
        """
        Verifies if the argument of a function has a type annotation

        :param arg: the python ast arg node
        """
        if arg.annotation is None:
            self._log_error(
                CompilerError.TypeHintMissing(arg.lineno, arg.col_offset, symbol_id=arg.arg)
            )

        # continue to walk through the tree
        self.generic_visit(arg)

    def visit_Return(self, ret: ast.Return):
        """
        Verifies if the return of the function is the same type as the return type annotation

        :param ret: the python ast return node
        """
        ret_value: Any = self.visit(ret.value) if ret.value is not None else None
        if ret.value is not None and self.get_type(ret.value) is not Type.none:
            # multiple returns are not allowed
            if isinstance(ret.value, ast.Tuple):
                self._log_error(
                    CompilerError.TooManyReturns(ret.lineno, ret.col_offset)
                )
                return
            # it is returning something, but there is no type hint for return
            elif self._current_method.return_type is Type.none:
                self._log_error(
                    CompilerError.TypeHintMissing(ret.lineno, ret.col_offset, symbol_id=self._current_method_id)
                )
                return
        self.validate_type_variable_assign(ret, ret_value, self._current_method)

        # continue to walk through the tree
        self.generic_visit(ret)

    def _validate_return(self, node: ast.AST):
        if (self._current_method.return_type is not Type.none
                and hasattr(node, 'body')
                and not self._has_return(node)):
            lst = list(self.symbols.items())
            keys, values = [key_value[0] for key_value in lst], [key_value[-1] for key_value in lst]
            self._log_error(
                CompilerError.MissingReturnStatement(
                    line=node.lineno, col=node.col_offset,
                    symbol_id=keys[values.index(self._current_method)]
                )
            )

    def _has_return(self, node: ast.AST) -> bool:
        if not hasattr(node, 'body'):
            return False

        has_else_stmt: bool = hasattr(node, 'orelse')
        has_body_return: bool = any(isinstance(stmt, (ast.Return, ast.Pass)) for stmt in node.body)
        if has_body_return and not has_else_stmt:
            # if any statement in the function body is a return, all flow has a return
            return True

        body = [stmt for stmt in node.body if hasattr(stmt, 'body')]
        body_has_inner_return: bool = len(body) > 0 and all(self._has_return(stmt) for stmt in body)
        if not has_body_return and not body_has_inner_return:
            # for and while nodes must to check if there is a return inside the else statement
            if not isinstance(node, (ast.For, ast.AsyncFor, ast.While)):
                return False

        if has_else_stmt:
            if any(isinstance(stmt, (ast.Return, ast.Pass)) for stmt in node.orelse):
                return True
            else:
                orelse = [stmt for stmt in node.orelse if hasattr(stmt, 'body')]
                return len(orelse) > 0 and all(self._has_return(stmt) for stmt in orelse)
        return body_has_inner_return

    def _check_base_init_call(self, node: ast.AST):
        if not self._current_method.is_init or not isinstance(self._current_class, ClassType):
            # if the method is not an user class __init__, don't check
            return

        if len(self._current_class.bases) == 0:
            # nothing to check if class has no bases
            return

        if len(self._super_calls) == 0:
            self._log_error(CompilerError.MissingInitCall(line=node.lineno,
                                                          col=node.col_offset))
        else:
            self._super_calls.clear()

    def visit_Assign(self, assign: ast.Assign):
        """
        Verifies if it is a multiple assignments statement

        :param assign: the python ast variable assignment node
        """
        for target in assign.targets:
            self.validate_type_variable_assign(target, assign.value)

    def visit_AnnAssign(self, ann_assign: ast.AnnAssign):
        """
        Verifies if the assigned type is the same as the variable type

        :param ann_assign: the python ast variable annotated assignment node
        """
        # if value is None, it is a declaration
        if ann_assign.value is not None:
            self.validate_type_variable_assign(ann_assign.target, ann_assign.value, implicit_cast=True)

    def visit_AugAssign(self, aug_assign: ast.AugAssign):
        """
        Verifies if the types of the target variable and the value are compatible with the operation

        If the operation is valid, changes de Python operator by the Boa operator in the syntax tree

        :param aug_assign: the python ast augmented assignment node
        """
        operation = self.validate_binary_operation(aug_assign, aug_assign.target, aug_assign.value)
        if operation is not None:
            self.validate_type_variable_assign(aug_assign.target, operation)
            aug_assign.op = operation

    def validate_type_variable_assign(self, node: ast.AST, value: Any, target: Any = None,
                                      implicit_cast: bool = False) -> bool:
        value_type: IType = self.get_type(value)
        if isinstance(value, ast.AST):
            value = self.visit(value)
            if isinstance(value, ast.Name):
                symbol: ISymbol = self.get_symbol(value.id)
                if isinstance(symbol, IExpression):
                    value_type = symbol.type
                elif isinstance(symbol, IType):
                    value_type = symbol
                else:
                    self._log_error(
                        CompilerError.UnresolvedReference(
                            value.lineno, value.col_offset,
                            symbol_id=value.id
                        ))

        if target is not None:
            target_type = self.get_type(target)
        else:
            target_type = None
            var: ISymbol = self.get_symbol(node.id if hasattr(node, 'id') else node)
            if isinstance(var, Variable):
                if (var.type is UndefinedType
                        or (var.type is Undefined and var not in self.symbols.values())):
                    var = var.copy()

                if var.type in (None, UndefinedType, Undefined):
                    # it is an declaration with assignment and the value is neither literal nor another variable
                    var.set_type(value_type)
                target_type = var.type

            elif isinstance(node, ast.Name):
                self._log_error(
                    CompilerError.UnresolvedReference(
                        node.lineno, node.col_offset,
                        symbol_id=node.id
                    ))
                return False

        if self._current_scope is not None:
            if isinstance(node, ast.Name):
                if node.id not in self._current_scope:
                    if implicit_cast:
                        value_type = target_type
                    elif not isinstance(value_type, Collection):
                        target_type = value_type
                    elif not (isinstance(type(value_type), type(target_type)) and
                              (value_type.value_type is Type.any or value_type.valid_key is Type.any)):
                        # if the collection is generic, let the validation for the outer scope
                        value_type = target_type

                    self._current_scope.include_symbol(node.id, Variable(value_type))

                else:
                    if self._current_method is not None and node.id in self._current_method.symbols:
                        can_change_target_type = self._current_method.symbols[node.id].type is UndefinedType
                    else:
                        can_change_target_type = True

                    if can_change_target_type:
                        if (not target_type.is_type_of(value_type) and
                                value != target_type.default_value):
                            target_type = value_type
                        self._current_scope.include_symbol(node.id, Variable(value_type))

        if not target_type.is_type_of(value_type) and value != target_type.default_value:
            if not implicit_cast:
                self._log_error(
                    CompilerError.MismatchedTypes(
                        node.lineno, node.col_offset,
                        actual_type_id=value_type.identifier,
                        expected_type_id=target_type.identifier
                    ))
                return False
            else:
                self._log_warning(
                    CompilerWarning.TypeCasting(
                        node.lineno, node.col_offset,
                        origin_type_id=value_type.identifier,
                        cast_type_id=target_type.identifier
                    )
                )

        return True

    def visit_Subscript(self, subscript: ast.Subscript) -> IType:
        """
        Verifies if the subscribed value is a sequence

        :param subscript: the python ast subscript node
        :return: the type of the accessed value if it is valid. Type.none otherwise.
        """
        if isinstance(subscript.slice, ast.Slice):
            return self.validate_slice(subscript, subscript.slice)

        return self.validate_get_or_set(subscript, subscript.slice)

    def validate_get_or_set(self, subscript: ast.Subscript, index_node: ast.AST) -> IType:
        """
        Verifies if the subscribed value is a sequence and if the index is valid to this sequence

        :param subscript: the python ast subscript node
        :param index_node: the subscript index
        :return: the type of the accessed value if it is valid. Type.none otherwise.
        """
        is_internal = hasattr(subscript, 'is_internal_call') and subscript.is_internal_call

        value = self.visit(subscript.value)
        index = self.visit(index_node)

        if isinstance(value, ast.Name):
            value = self.get_symbol(value.id, is_internal=is_internal)
        if isinstance(index, ast.Name):
            index = self.get_symbol(index.id, is_internal=is_internal)
        if not isinstance(index, tuple):
            index = (index,)

        # if it is a type hint, returns the outer type
        if isinstance(value, IType) and all(isinstance(i, IType) for i in index):
            if isinstance(value, Collection):
                value = value.build_collection(*index)

            value_to_be_built = index if isinstance(value, MetaType) else value
            if not isinstance(value_to_be_built, tuple):
                value_to_be_built = (value_to_be_built,)
            return TypeUtils.type.build(*value_to_be_built)

        symbol_type: IType = self.get_type(value)
        index_type: IType = self.get_type(index[0])

        # only sequence types can be subscribed
        if not isinstance(symbol_type, Collection):
            self._log_error(
                CompilerError.UnresolvedOperation(
                    subscript.lineno, subscript.col_offset,
                    type_id=symbol_type.identifier,
                    operation_id=Operator.Subscript)
            )
            return symbol_type
        # the sequence can't use the given type as index
        if not symbol_type.is_valid_key(index_type):
            self._log_error(
                CompilerError.MismatchedTypes(
                    subscript.lineno, subscript.col_offset,
                    actual_type_id=index_type.identifier,
                    expected_type_id=symbol_type.valid_key.identifier)
            )
        # it is setting a value in a sequence that doesn't allow reassign values
        elif isinstance(subscript.ctx, ast.Store) and not symbol_type.can_reassign_values:
            self._log_error(
                CompilerError.UnresolvedOperation(
                    subscript.lineno, subscript.col_offset,
                    type_id=symbol_type.identifier,
                    operation_id=Operator.Subscript)
            )
        return symbol_type.item_type

    def validate_slice(self, subscript: ast.Subscript, slice_node: ast.Slice) -> IType:
        """
        Verifies if the subscribed value is a sequence and if the slice is valid to this sequence

        :param subscript: the python ast subscript node
        :param slice_node: the subscript slice
        :return: the type of the accessed value if it is valid. Type.none otherwise.
        """
        value = self.visit(subscript.value)
        lower, upper, step = (self.get_type(value) for value in self.visit(slice_node))

        symbol_type: IType = self.get_type(value)
        # only collection types can be subscribed
        if not isinstance(symbol_type, Collection):
            self._log_error(
                CompilerError.UnresolvedOperation(
                    subscript.lineno, subscript.col_offset,
                    type_id=symbol_type.identifier,
                    operation_id=Operator.Subscript)
            )
            return Type.none

        lower = lower if lower is not Type.none else symbol_type.valid_key
        upper = upper if upper is not Type.none else symbol_type.valid_key
        step = step if step is not Type.none else symbol_type.valid_key

        # TODO: remove when slices of other sequence types are implemented
        if (not symbol_type.is_valid_key(lower)
                or not symbol_type.is_valid_key(upper)
                or (step is not Type.none and not symbol_type.is_valid_key(step))
            ):
            actual: Tuple[IType, ...] = (lower, upper) if step is Type.none else (lower, upper, step)
            self._log_error(
                CompilerError.MismatchedTypes(
                    subscript.lineno, subscript.col_offset,
                    expected_type_id=[symbol_type.valid_key.identifier for value in actual],
                    actual_type_id=[value.identifier for value in actual]
                )
            )
        else:
            return symbol_type

        return Type.none

    def visit_While(self, while_node: ast.While):
        """
        Verifies if the type of while test is valid

        :param while_node: the python ast while statement node
        """
        test = self.visit(while_node.test)
        test_type: IType = self.get_type(test)

        if test_type is not Type.bool:
            self._log_error(
                CompilerError.MismatchedTypes(
                    while_node.lineno, while_node.col_offset,
                    actual_type_id=test_type.identifier,
                    expected_type_id=Type.bool.identifier)
            )

        # continue to walk through the tree
        for stmt in while_node.body:
            self.visit(stmt)
        for stmt in while_node.orelse:
            self.visit(stmt)

    def visit_For(self, for_node: ast.For):
        """
        Verifies if the type of for iterator is valid

        :param for_node: the python ast for node
        """
        iterator = self.visit(for_node.iter)
        iterator_type: IType = self.get_type(iterator)

        if not isinstance(iterator_type, Collection):
            self._log_error(
                CompilerError.MismatchedTypes(
                    for_node.lineno, for_node.col_offset,
                    actual_type_id=iterator_type.identifier,
                    expected_type_id=Type.sequence.identifier)
            )

        # TODO: change when optimizing for loops
        elif self.get_type(for_node.target) != iterator_type.item_type:
            target_id = self.visit(for_node.target)
            if isinstance(target_id, ast.Name):
                target_id = target_id.id
            symbol = self.get_symbol(target_id)
            symbol.set_type(iterator_type.item_type)

        # continue to walk through the tree
        for stmt in for_node.body:
            self.visit(stmt)
        for stmt in for_node.orelse:
            self.visit(stmt)

    def visit_If(self, if_node: ast.If):
        """
        Verifies if the type of if test is valid

        :param if_node: the python ast if statement node
        """
        is_instance_if, is_instance_else = self.validate_if(if_node)

        # continue to walk through the tree
        self.new_local_scope(is_instance_if)
        for stmt in if_node.body:
            self.visit(stmt)
        is_instance_if = self.pop_local_scope().symbols

        self.new_local_scope(is_instance_else)
        for stmt in if_node.orelse:
            self.visit(stmt)
        is_instance_else = self.pop_local_scope().symbols

        last_scope = self._current_scope if len(self._scope_stack) > 0 else self._current_method
        # updates the outer scope for each variable that is changed in both branches
        intersected_ids = is_instance_if.keys() & is_instance_else.keys()
        for changed_symbol in intersected_ids:
            new_value_type_if_branch = self.get_type(is_instance_if[changed_symbol])
            new_value_type_else_branch = self.get_type(is_instance_else[changed_symbol])
            new_type = Type.union.build([new_value_type_if_branch, new_value_type_else_branch])
            last_scope.include_symbol(changed_symbol, Variable(new_type))

        # updates with the variables assigned in a branch that doesn't exist in the outer scope
        self._include_symbols_to_scope(last_scope, is_instance_if, intersected_ids)
        self._include_symbols_to_scope(last_scope, is_instance_else, intersected_ids)

    def _include_symbols_to_scope(self, scope: SymbolScope, other_scope: Dict[str, ISymbol], items_filter: Set[str]):
        for new_symbol_id in {key for key in other_scope if key not in items_filter}:
            new_symbol = other_scope[new_symbol_id]
            new_value_type = self.get_type(new_symbol)
            outer_symbol = self.get_symbol(new_symbol_id)
            outer_value_type = outer_symbol.type if isinstance(outer_symbol, IExpression) else Type.none

            if (isinstance(outer_symbol, Variable) and isinstance(new_symbol, Variable)
                    and not new_symbol.is_reassigned):
                continue

            if isinstance(outer_symbol, IType):
                new_type = Type.union.build([new_value_type, outer_value_type])
            else:
                new_type = new_value_type
            scope.include_symbol(new_symbol_id, Variable(new_type))

    def visit_IfExp(self, if_node: ast.IfExp):
        """
        Verifies if the type of if test is valid

        :param if_node: the python ast if expression node
        """
        is_instance_if, is_instance_else = self.validate_if(if_node)
        body = if_node.body
        orelse = if_node.orelse
        if_value = body[-1] if isinstance(body, list) and len(body) > 0 else body
        else_value = orelse[-1] if isinstance(orelse, list) and len(orelse) > 0 else orelse

        self.new_local_scope(is_instance_if)
        if_type: IType = self.get_type(if_value)
        self.pop_local_scope()

        self.new_local_scope(is_instance_else)
        else_type: IType = self.get_type(else_value)
        self.pop_local_scope()

        return Type.get_generic_type(if_type, else_type)

    def validate_if(self, if_node: ast.AST) -> Tuple[Dict[str, ISymbol], Dict[str, ISymbol]]:
        """
        Verifies if the type of if test is valid

        :param if_node: the python ast if statement node
        :type if_node: ast.If or ast.IfExp
        """
        test = self.visit(if_node.test)
        test_type: IType = self.get_type(test)

        return self._get_is_instance_function_calls(if_node.test)

    def _get_is_instance_function_calls(self, node: ast.AST) -> Tuple[Dict[str, ISymbol], Dict[str, ISymbol]]:
        from boa3.model.builtin.method.isinstancemethod import IsInstanceMethod
        is_instance_objs = [x for x, symbol in self.symbols.items()
                            if isinstance(symbol, IsInstanceMethod)]
        is_instance_objs.append(isinstance.__name__)

        is_instance_symbols = {}
        is_not_instance_symbols = {}

        if isinstance(node, ast.BoolOp) and isinstance(node.op, type(BinaryOp.And)):
            conditions = node.values
        else:
            conditions = [node]

        for condition in conditions:
            negate = False
            if isinstance(condition, ast.UnaryOp) and isinstance(condition.op, type(UnaryOp.Not)):
                condition = condition.operand
                negate = True

            # verifies if condition is an is_instance condition
            is_instance_condition = (isinstance(condition, ast.Call) and isinstance(condition.func, ast.Name)
                                     and condition.func.id in is_instance_objs and len(condition.args) == 2)

            # verifies if condition is a identity condition (is None or is not None)
            identity_condition = (isinstance(condition, ast.Compare)
                                  and isinstance(condition.ops[0], (type(BinaryOp.IsNone), type(BinaryOp.IsNotNone))))

            if identity_condition and isinstance(condition.ops[0], type(BinaryOp.IsNotNone)):
                negate = True

            if is_instance_condition or identity_condition:
                if is_instance_condition:
                    left_value = condition.args[0]
                    right_value = condition.args[1]

                else:
                    left_value = condition.left
                    right_value = Type.none

                original = self.get_symbol(left_value)
                if isinstance(original, Variable) and isinstance(left_value, ast.Name):
                    original_id = left_value.id
                    original_type = (original.type
                                     if original_id not in is_instance_symbols
                                     else is_instance_symbols[original_id])

                    instance_obj = self.get_symbol(condition.func.id) if is_instance_condition else None
                    if isinstance(instance_obj, type(Builtin.IsInstance)):
                        is_instance_type = instance_obj._instances_type
                        if isinstance(is_instance_type, list):
                            is_instance_type = Type.union.build(is_instance_type)
                    else:
                        is_instance_type = self.get_type(right_value)
                    is_instance_type = is_instance_type.intersect_type(original_type)
                    negation = original_type.except_type(is_instance_type)
                    resulting_type = is_instance_type if not negate else negation

                    if original_id not in is_not_instance_symbols:
                        is_instance_symbols[original_id] = resulting_type
                    else:
                        is_instance_symbols[original_id] = is_instance_symbols[original_id].except_type(resulting_type)

                    if len(is_instance_symbols) == 1:
                        is_not_instance_symbols[original_id] = is_instance_type if negate else negation
                    elif len(is_not_instance_symbols) == 1:
                        # if there is more than one isinstance it's not possible to determine in compiler time what is
                        # the type of the variables on the else body
                        is_not_instance_symbols.clear()

        for key, value in is_instance_symbols.copy().items():
            if value == self.get_type(self.get_symbol(key)):
                is_instance_symbols.pop(key)
            else:
                is_instance_symbols[key] = Variable(value)

        for key, value in is_not_instance_symbols.copy().items():
            if value == self.get_type(self.get_symbol(key)):
                is_not_instance_symbols.pop(key)
            else:
                is_not_instance_symbols[key] = Variable(value)
        return is_instance_symbols, is_not_instance_symbols

    def visit_BinOp(self, bin_op: ast.BinOp) -> Optional[IType]:
        """
        Verifies if the types of the operands are valid to the operation

        If the operation is valid, changes de Python operator by the Boa operator in the syntax tree

        :param bin_op: the python ast binary operation node
        :return: the type of the result of the operation if the operation is valid. Otherwise, returns None
        :rtype: IType or None
        """
        operation = self.validate_binary_operation(bin_op, bin_op.left, bin_op.right)
        if operation is not None:
            bin_op.op = operation
            return operation.result

    def validate_binary_operation(self, node: ast.AST, left_op: ast.AST, right_op: ast.AST) -> Optional[IOperation]:
        """
        Validates a ast node that represents a binary operation

        :param node: ast node that represents a binary operation
        :param left_op: ast node that represents the left operand of the operation
        :param right_op: ast node that represents the right operand of the operation
        :return: the corresponding :class:`BinaryOperation` if is valid. None otherwise.
        """
        if not hasattr(node, 'op'):
            return

        operator: Operator = self.get_operator(node.op)
        l_operand = self.visit(left_op)
        r_operand = self.visit(right_op)

        if not isinstance(operator, Operator):
            # the operator is invalid or it was not implemented yet
            self._log_error(
                CompilerError.UnresolvedReference(node.lineno, node.col_offset, type(node.op).__name__)
            )

        try:
            operation: IOperation = self.get_bin_op(operator, r_operand, l_operand)
            if operation is None:
                self._log_error(
                    CompilerError.NotSupportedOperation(node.lineno, node.col_offset, operator)
                )
            elif not operation.is_supported:
                # number float division is not supported by Neo VM
                self._log_error(
                    CompilerError.NotSupportedOperation(node.lineno, node.col_offset, operator)
                )
            else:
                return operation
        except CompilerError.MismatchedTypes as raised_error:
            raised_error.line = node.lineno
            raised_error.col = node.col_offset
            # raises the exception with the line/col info
            self._log_error(raised_error)

    def get_bin_op(self, operator: Operator, right: Any, left: Any) -> IOperation:
        """
        Returns the binary operation specified by the operator and the types of the operands

        :param operator: the operator
        :param right: right operand
        :param left: left operand

        :return: Returns the corresponding :class:`BinaryOperation` if the types are valid.
        :raise MismatchedTypes: raised if the types aren't valid for the operator
        """
        l_type: IType = self.get_type(left)
        r_type: IType = self.get_type(right)

        if l_type is None or r_type is None:
            return BinaryOp.get_operation_by_operator(operator, l_type if l_type is not None else r_type)

        actual_types = (l_type.identifier, r_type.identifier)
        operation: IOperation = BinaryOp.validate_type(operator, l_type, r_type)

        if operation is not None:
            return operation
        else:
            if operator.requires_right_operand_for_validation():
                left = l_type
                right = r_type
            else:
                left = l_type if left is not None else r_type
                right = None

            expected_op: BinaryOperation = BinaryOp.get_operation_by_operator(operator, left, right)
            expected_types = (expected_op.left_type.identifier, expected_op.right_type.identifier)
            raise CompilerError.MismatchedTypes(0, 0, expected_types, actual_types)

    def visit_UnaryOp(self, un_op: ast.UnaryOp) -> Optional[IType]:
        """
        Verifies if the type of the operand is valid to the operation

        If the operation is valid, changes de Python operator by the Boa operator in the syntax tree

        :param un_op: the python ast unary operation node
        :return: the type of the result of the operation if the operation is valid. Otherwise, returns None
        :rtype: IType or None
        """
        operator: Operator = self.get_operator(un_op.op)
        operand = self.visit(un_op.operand)

        if not isinstance(operator, Operator):
            # the operator is invalid or it was not implemented yet
            self._log_error(
                CompilerError.UnresolvedReference(un_op.lineno, un_op.col_offset, type(un_op.op).__name__)
            )

        try:
            operation: UnaryOperation = self.get_un_op(operator, operand)
            if operation is None:
                self._log_error(
                    CompilerError.NotSupportedOperation(un_op.lineno, un_op.col_offset, operator)
                )
            elif not operation.is_supported:
                self._log_error(
                    CompilerError.NotSupportedOperation(un_op.lineno, un_op.col_offset, operator)
                )
            else:
                un_op.op = operation
                return operation.result
        except CompilerError.MismatchedTypes as raised_error:
            raised_error.line = un_op.lineno
            raised_error.col = un_op.col_offset
            # raises the exception with the line/col info
            self._log_error(raised_error)

    def get_un_op(self, operator: Operator, operand: Any) -> UnaryOperation:
        """
        Returns the binary operation specified by the operator and the types of the operands

        :param operator: the operator
        :param operand: the operand

        :return: Returns the corresponding :class:`UnaryOperation` if the types are valid.
        :raise MismatchedTypes: raised if the types aren't valid for the operator
        """
        op_type: IType = self.get_type(operand)

        actual_type: str = op_type.identifier
        operation: UnaryOperation = UnaryOp.validate_type(operator, op_type)

        if operation is not None:
            return operation
        else:
            expected_op: UnaryOperation = UnaryOp.get_operation_by_operator(operator)
            expected_type: str = expected_op.operand_type.identifier
            raise CompilerError.MismatchedTypes(0, 0, expected_type, actual_type)

    def visit_Assert(self, assert_: ast.Assert):
        """
        Verifies if the types of condition and error message in the assert are valid

        If the operations are valid, changes de Python operator by the Boa operator in the syntax tree

        :param assert_: the python ast assert operation node
        :return: the type of the result of the operation if the operation is valid. Otherwise, returns None
        """
        self.visit(assert_.test)

        if assert_.msg is not None:
            msg = self.visit(assert_.msg)
            msg_type = self.get_type(msg)

            if not Type.str.is_type_of(msg_type) and not Type.bytes.is_type_of(msg_type):

                # TODO: remove this error when str constructor is implemented
                self._log_error(
                    CompilerError.MismatchedTypes(
                        assert_.msg.lineno, assert_.msg.col_offset,
                        expected_type_id=Type.str.identifier,
                        actual_type_id=str(msg_type)
                    )
                )

    def visit_Compare(self, compare: ast.Compare) -> Optional[IType]:
        """
        Verifies if the types of the operands are valid to the compare operations

        If the operations are valid, changes de Python operator by the Boa operator in the syntax tree

        :param compare: the python ast compare operation node
        :return: the type of the result of the operation if the operation is valid. Otherwise, returns None
        :rtype: IType or None
        """
        if len(compare.comparators) != len(compare.ops):
            self._log_error(
                CompilerError.IncorrectNumberOfOperands(
                    compare.lineno,
                    compare.col_offset,
                    len(compare.ops),
                    len(compare.comparators) + 1  # num comparators + compare.left
                )
            )

        line = compare.lineno
        col = compare.col_offset
        try:
            return_type = None
            l_operand = self.visit_value(compare.left)
            for index, op in enumerate(compare.ops):
                operator: Operator = self.get_operator(op)
                r_operand = self.visit_value(compare.comparators[index])

                if not isinstance(operator, Operator):
                    # the operator is invalid or it was not implemented yet
                    self._log_error(
                        CompilerError.UnresolvedReference(line, col, type(op).__name__)
                    )

                operation: IOperation = self.get_bin_op(operator, r_operand, l_operand)
                if operation is None:
                    self._log_error(
                        CompilerError.NotSupportedOperation(line, col, operator)
                    )
                elif not operation.is_supported:
                    self._log_error(
                        CompilerError.NotSupportedOperation(line, col, operator)
                    )
                else:
                    compare.ops[index] = operation
                    return_type = operation.result

                line = compare.comparators[index].lineno
                col = compare.comparators[index].col_offset
                l_operand = r_operand

            return return_type
        except CompilerError.MismatchedTypes as raised_error:
            raised_error.line = line
            raised_error.col = col
            # raises the exception with the line/col info
            self._log_error(raised_error)

    def visit_BoolOp(self, bool_op: ast.BoolOp) -> Optional[IType]:
        """
        Verifies if the types of the operands are valid to the boolean operations

        If the operations are valid, changes de Python operator by the Boa operator in the syntax tree

        :param bool_op: the python ast boolean operation node
        :return: the type of the result of the operation if the operation is valid. Otherwise, returns None
        :rtype: IType or None
        """
        lineno: int = bool_op.lineno
        col_offset: int = bool_op.col_offset
        try:
            return_type: IType = None
            bool_operation: IOperation = None
            operator: Operator = self.get_operator(bool_op.op)

            if not isinstance(operator, Operator):
                # the operator is invalid or it was not implemented yet
                self._log_error(
                    CompilerError.UnresolvedReference(lineno, col_offset, type(operator).__name__)
                )

            l_operand = self.visit(bool_op.values[0])
            for index, operand in enumerate(bool_op.values[1:]):
                r_operand = self.visit(operand)

                operation: IOperation = self.get_bin_op(operator, r_operand, l_operand)
                if operation is None:
                    self._log_error(
                        CompilerError.NotSupportedOperation(lineno, col_offset, operator)
                    )
                elif bool_operation is None:
                    return_type = operation.result
                    bool_operation = operation

                lineno = operand.lineno
                col_offset = operand.col_offset
                l_operand = r_operand

            bool_op.op = bool_operation
            return return_type
        except CompilerError.MismatchedTypes as raised_error:
            raised_error.line = lineno
            raised_error.col = col_offset
            # raises the exception with the line/col info
            self._log_error(raised_error)

    def get_operator(self, node: Union[ast.operator, Operator, IOperation]) -> Optional[Operator]:
        """
        Gets the :class:`Operator` equivalent to given Python ast operator

        :param node: object with the operator data
        :type node: ast.operator or Operator or IOperation
        :return: the Boa operator equivalent to the node. None if it doesn't exit.
        :rtype: Operator or None
        """
        # the node has already been visited
        if isinstance(node, Operator):
            return node
        elif isinstance(node, IOperation):
            return node.operator

        return Operator.get_operation(node)

    def visit_Call(self, call: ast.Call):
        """
        Verifies if the number of arguments is correct

        :param call: the python ast function call node
        :return: the result type of the called function
        """
        if isinstance(call.func, ast.Name):
            callable_id: str = call.func.id
            is_internal = hasattr(call, 'is_internal_call') and call.is_internal_call
            callable_target = self.get_symbol(callable_id, is_internal)
        else:
            callable_id, callable_target = self.get_callable_and_update_args(call)  # type: str, ISymbol

        callable_target = self.validate_builtin_callable(callable_id, callable_target, call.args)

        if not isinstance(callable_target, Callable):
            # if the outer call is a builtin, enable call even without the import
            builtin_symbol = Builtin.get_symbol(callable_id)
            if builtin_symbol is not None:
                callable_target = builtin_symbol

        callable_method_id = None
        if isinstance(callable_target, ClassType):
            callable_target = callable_target.constructor_method()
            callable_method_id = constants.INIT_METHOD_ID

        if not isinstance(callable_target, Callable):
            # the symbol doesn't exists or is not a function
            # if it is None, the error was already logged
            if callable_id is not None:
                if callable_method_id is not None:
                    callable_id = '{0}.{1}()'.format(callable_id, callable_method_id)
                self._log_error(
                    CompilerError.UnresolvedReference(call.func.lineno, call.func.col_offset, callable_id)
                )
        else:
            if not self.check_call_scope(call, callable_target, callable_id):
                # errors are logged and handled by the method itself
                return self.get_type(callable_target)

            if callable_target is Builtin.NewEvent:
                return callable_target.return_type

            private_identifier = None  # used for validating internal builtin methods
            if self.validate_callable_arguments(call, callable_target):
                args = [self.get_type(param, use_metatype=True) for param in call.args]
                if isinstance(callable_target, IBuiltinMethod):
                    # if the arguments are not generic, build the specified method
                    if callable_target.raw_identifier.startswith('-'):
                        private_identifier = callable_target.raw_identifier
                    callable_target: IBuiltinMethod = callable_target.build(args)
                    if not callable_target.is_supported:
                        self._log_error(
                            CompilerError.NotSupportedOperation(call.lineno, call.col_offset,
                                                                callable_target.not_supported_str(callable_id))
                        )
                        return callable_target.return_type

                    if callable_target.is_cast:
                        # every typing cast raises a warning
                        cast_types = callable_target.cast_types
                        if cast_types is None:
                            origin_type_id = 'unknown'
                            cast_type_id = callable_target.type.identifier
                        else:
                            origin_type_id = cast_types[0].identifier
                            cast_type_id = cast_types[1].identifier

                        if not hasattr(call, 'is_internal_call') or not call.is_internal_call:
                            self._log_warning(
                                CompilerWarning.TypeCasting(call.lineno, call.col_offset,
                                                            origin_type_id=origin_type_id,
                                                            cast_type_id=cast_type_id)
                            )

                self.validate_passed_arguments(call, args, callable_id, callable_target)

            if private_identifier is not None and callable_target.identifier != private_identifier:
                # updates the callable_id for validation of internal builtin that accepts generic variables
                callable_id = private_identifier
            self.update_callable_after_validation(call, callable_id, callable_target)

        return self.get_type(callable_target)

    def get_callable_and_update_args(self, call: ast.Call) -> Tuple[str, ISymbol]:
        attr: Attribute = self.visit(call.func)
        if not isinstance(attr, Attribute):
            attr_id = str(attr)
            attr_id_split = attr_id.split(constants.ATTRIBUTE_NAME_SEPARATOR)
            attr_call_id = attr_id
            if len(attr_id_split) > 0:
                attr_call_id = constants.ATTRIBUTE_NAME_SEPARATOR.join(attr_id_split[:-1])

            self._log_error(
                CompilerError.UnresolvedReference(call.func.lineno, call.func.col_offset, attr_call_id)
            )

            return attr_id, None

        arg0, callable_target, callable_id = attr.values

        if isinstance(arg0, Package):
            # visit works only with ast classes
            package = arg0
            package_symbol = self.get_symbol(package.identifier, check_raw_id=True)

            while package_symbol is None and package.parent is not None:
                package = package.parent
                package_symbol = self.get_symbol(package.identifier, check_raw_id=True)

            if package_symbol is None:
                arg0_identifier = package.identifier
            elif isinstance(package_symbol, Package):
                arg0_identifier = package_symbol
            else:
                arg0_identifier = package

        elif isinstance(arg0, UserClass):
            arg0_identifier = arg0.identifier
        elif isinstance(arg0, ast.AST):
            arg0_identifier = self.visit(arg0)
        else:
            arg0_identifier = None

        if isinstance(arg0_identifier, ast.Name):
            arg0_identifier = arg0_identifier.id

        if (callable_target is not None
                and not isinstance(self.get_symbol(arg0_identifier), (IType, Import, Package))
                and not isinstance(arg0_identifier, UserClass)
                and (len(call.args) < 1 or call.args[0] != arg0)):
            # move self to the arguments
            # don't move if it's class method
            if not (isinstance(arg0, ClassType) and callable_id in arg0.class_methods):
                call.args.insert(0, arg0)

        if len(call.args) > 0 and isinstance(callable_target, IBuiltinMethod) and callable_target.has_self_argument:
            self_type: IType = self.get_type(call.args[0])
            caller = self.get_symbol(arg0_identifier)
            if isinstance(caller, IType) and not caller.is_type_of(self_type):
                self_type = caller
            callable_target = callable_target.build(self_type)

        return callable_id, callable_target

    def check_call_scope(self, call: ast.Call, callable: Callable, callable_id: str):
        error_count = len(self.errors)
        if not isinstance(call.func, ast.Attribute):
            # if the call doesn't came from an attribute, there's nothing to be validated
            return True

        attr_node: ast.Attribute = call.func
        if isinstance(attr_node.value, ast.Name):
            attribute_symbol = self.get_symbol(attr_node.value.id)
        else:
            attribute_symbol = self.get_symbol(attr_node.value)
        is_from_type_name = isinstance(attribute_symbol, IType)

        attribute_type = self.get_type(attribute_symbol)
        if not isinstance(attribute_type, UserClass):
            # TODO: change when class specific scopes are implemented in the built-ins
            return True

        # TODO: remove this verification when calling an instance function from a class is implemented
        if is_from_type_name and isinstance(callable, Method) and hasattr(attribute_symbol, 'instance_methods') \
                and callable_id in attribute_symbol.instance_methods:
            callable_complete_id = f'{attribute_type.identifier}.{callable_id}'
            self._log_error(
                CompilerError.NotSupportedOperation(call.func.lineno, call.func.col_offset, callable_complete_id)
            )
        elif is_from_type_name and callable_id not in attribute_type.class_symbols:
            # the current symbol doesn't exist in the class scope
            callable_complete_id = f'{attribute_type.identifier}.{callable_id}'
            self._log_error(
                CompilerError.UnresolvedReference(call.func.lineno, call.func.col_offset, callable_complete_id)
            )
        elif not is_from_type_name and callable_id not in attribute_type.instance_symbols:
            # the current symbol doesn't exist in the instance scope
            callable_complete_id = f'{attribute_type.identifier}().{callable_id}'
            self._log_error(
                CompilerError.UnresolvedReference(call.func.lineno, call.func.col_offset, callable_complete_id)
            )

        return len(self.errors) == error_count

    def validate_builtin_callable(self, callable_id: str, callable_target: ISymbol,
                                  call_args: List[ast.AST] = None) -> ISymbol:
        if call_args is None:
            call_args = []

        if not isinstance(callable_target, Callable):
            # verify if it is a builtin method with its name shadowed
            call_target = Builtin.get_symbol(callable_id)
            if not isinstance(call_target, Callable) and self.is_exception(callable_id):
                call_target = Builtin.Exception

            callable_target = call_target if call_target is not None else callable_target
        if isinstance(callable_target, IBuiltinMethod):
            # verify if it's a variation of the default builtin method
            args = [self.get_type(param, use_metatype=True) for param in call_args]

            from boa3.model.builtin.method import SuperMethod
            # TODO: change when implementing super() with args
            if (isinstance(callable_target, SuperMethod)
                    and isinstance(self._current_method, Method) and self._current_method.has_cls_or_self):
                args.insert(0, self._current_class)

            new_target = callable_target.build(args)
            if new_target is not None:
                callable_target = new_target

            if isinstance(callable_target, SuperMethod) and callable_target not in self._super_calls:
                self._super_calls.append(callable_target)
        return callable_target

    def validate_callable_arguments(self, call: ast.Call, callable_target: Callable) -> bool:
        if callable_target.has_starred_argument and not hasattr(call, 'checked_starred_args'):

            if (len(call.args) >= len(callable_target.args)
                    and (len(call.args) == 0 or not isinstance(call.args[0], ast.Starred))):
                # starred argument is always the last argument
                len_args_without_starred = len(callable_target.args) - 1
                args = self.parse_to_node(str(Type.tuple.default_value), call)

                # include the arguments into a tuple to be assigned to the starred argument
                args.elts = call.args[len_args_without_starred:]
                call.args[len_args_without_starred:] = [args]

            call.checked_starred_args = True

        ignore_first_argument = int(callable_target.has_cls_or_self)  # 1 if True, 0 otherwise
        len_call_args = len(call.args)
        len_call_keywords = len(call.keywords)
        callable_required_args = len(callable_target.args_without_default) - ignore_first_argument

        # verifies if a non-default arg is being called as a keyword
        necessary_kwargs = list(callable_target.args_without_default.keys())[len_call_args:callable_required_args]
        keywords_names_used = []
        all_required_arg_have_values = True
        for keyword in call.keywords:
            keywords_names_used.append(keyword.arg)
        index = 0
        while index < len(necessary_kwargs) and all_required_arg_have_values:
            if necessary_kwargs[index] not in keywords_names_used:
                all_required_arg_have_values = False
            index += 1

        # verifies if a kwarg is being used but is not an argument for the function
        index = 0
        unexpected_kwarg = None
        while unexpected_kwarg is None and index < len(call.keywords):
            if call.keywords[index].arg not in list(callable_target.args.keys()):
                unexpected_kwarg = call.keywords[index].value
            index += 1

        # verifies if a kwarg that was already used as a positional argument is being used again
        implicit_args = list(callable_target.args_without_default.keys())[:len_call_args]
        index = 0
        already_called_arg = None
        kwargs_used_names = []
        for keyword in call.keywords:
            kwargs_used_names.append(keyword.arg)
        while already_called_arg is None and index < len(implicit_args):
            if implicit_args[index] in kwargs_used_names:
                already_called_arg = call.keywords[index].value
            index += 1

        if len_call_args > len(callable_target.args) or unexpected_kwarg is not None or already_called_arg is not None:
            if unexpected_kwarg is not None:
                unexpected_arg = unexpected_kwarg
            elif already_called_arg is not None:
                unexpected_arg = already_called_arg
            else:
                unexpected_arg = call.args[len(callable_target.args) + ignore_first_argument]
            self._log_error(
                CompilerError.UnexpectedArgument(unexpected_arg.lineno, unexpected_arg.col_offset)
            )
            return False
        elif len_call_args + len_call_keywords < callable_required_args or not all_required_arg_have_values:
            missed_arg = list(callable_target.args)[len(call.args) + ignore_first_argument]
            self._log_error(
                CompilerError.UnfilledArgument(call.lineno, call.col_offset, missed_arg)
            )
            return False

        if isinstance(callable_target, IBuiltinMethod) and callable_target.requires_reordering:
            if not hasattr(call, 'was_reordered') or not call.was_reordered:
                callable_target.reorder(call.args)
                call.was_reordered = True

        if callable_required_args <= len_call_args < len(callable_target.args):
            included_args = len_call_args - callable_required_args
            for default in callable_target.defaults[included_args:]:
                call.args.append(self.clone(default))
        return True

    def validate_passed_arguments(self, call: ast.Call, args_types: List[IType], callable_id: str, callable: Callable):
        if isinstance(callable, IBuiltinMethod):
            builtin_analyser = BuiltinFunctionCallAnalyser(self, call, callable_id, callable, self._log)
            if builtin_analyser.validate():
                self.errors.extend(builtin_analyser.errors)
                self.warnings.extend(builtin_analyser.warnings)
                return

        param_types = []
        ignore_first_argument = int(callable.has_cls_or_self)  # 1 if True, 0 otherwise
        is_first_arg_cls_of_self = ignore_first_argument and len(call.args) == len(callable.args)

        # validate positional parameters
        if is_first_arg_cls_of_self:
            (self_id, self_value) = list(callable.args.items())[0]
            param_type = self._validate_argument_type(call.args[0], self_value.type, use_metatype=True)
            param_types.append(param_type)

        for index, param in enumerate(call.args[is_first_arg_cls_of_self:]):
            (arg_id, arg_value) = list(callable.args.items())[ignore_first_argument + index]

            param_type = self._validate_argument_type(param, arg_value.type, use_metatype=True)
            param_types.append(param_type)

        # validate keyword arguments
        for param in call.keywords:
            arg_value = callable.args[param.arg]
            param = param.value
            param_type = self._validate_argument_type(param, arg_value.type)
            param_types.append(param_type)

    def _validate_argument_type(self, param: ast.AST, arg_type: IType, use_metatype: bool = False) -> Optional[IType]:
        param_type = self.get_type(param, use_metatype=use_metatype)
        if arg_type.is_type_of(param_type):
            return param_type
        else:
            self._log_error(
                CompilerError.MismatchedTypes(
                    param.lineno, param.col_offset,
                    arg_type.identifier,
                    param_type.identifier
                ))

        return None

    def update_callable_after_validation(self, call: ast.Call, callable_id: str, callable_target: Callable):
        callable_target.add_call_origin(call)

        # if the arguments are not generic, include the specified method in the symbol table
        if (isinstance(callable_target, IBuiltinMethod)
                and callable_target.identifier != callable_id
                and callable_target.raw_identifier == callable_id):
            if callable_target.identifier not in self.symbols:
                self.symbols[callable_target.identifier] = callable_target
            call.func = ast.Name(lineno=call.func.lineno, col_offset=call.func.col_offset,
                                 ctx=ast.Load(), id=callable_target.identifier)

    def visit_Raise(self, raise_node: ast.Raise):
        """
        Visitor of the raise node

        Verifies if the raised object is a exception

        :param raise_node: the python ast raise node
        """
        raised = self.visit(raise_node.exc)
        raised_type: IType = self.get_type(raised)
        if raised_type is not Type.exception:
            self._log_error(
                CompilerError.MismatchedTypes(
                    raise_node.lineno, raise_node.col_offset,
                    actual_type_id=raised_type.identifier,
                    expected_type_id=Type.exception.identifier
                ))

    def visit_Try(self, try_node: ast.Try):
        """
        Visitor of the try node

        :param try_node: the python ast try node
        """
        self.validate_except_handlers(try_node)

        for stmt in try_node.body:
            self.visit(stmt)

        for exception_handler in try_node.handlers:
            self.visit(exception_handler)

        for stmt in try_node.orelse:
            self.visit(stmt)

        for stmt in try_node.finalbody:
            self.visit(stmt)

    def validate_except_handlers(self, try_node: ast.Try):
        if len(try_node.handlers) > 1:
            exception_handlers: List[ast.ExceptHandler] = try_node.handlers.copy()
            general_exc_handler: ast.ExceptHandler = next(
                (handler
                 for handler in exception_handlers
                 if (handler.type is None or  # no specified exception or is BaseException
                     (isinstance(handler.type, ast.Name) and handler.type.id == BaseException.__name__)
                     )
                 ), exception_handlers[0]
            )

            try_node.handlers = [general_exc_handler]
            exception_handlers.remove(general_exc_handler)

            for handler in exception_handlers:
                warnings = len(self.warnings)
                self.visit(handler)
                if warnings == len(self.warnings):
                    self._log_using_specific_exception_warning(handler.type)

        if len(try_node.handlers) == 1:
            self.visit(try_node.handlers[0])

    def visit_ExceptHandler(self, node: ast.ExceptHandler):
        """
        Visitor of the try except node

        :param node: the python ast try except node
        """
        logged_errors = False

        exception_type: IType = self.get_type(node.type)
        if node.type is not None and exception_type is not Type.exception:
            self._log_error(
                CompilerError.MismatchedTypes(line=node.type.lineno,
                                              col=node.type.col_offset,
                                              expected_type_id=Type.exception.identifier,
                                              actual_type_id=exception_type.identifier)
            )
            logged_errors = True

        if node.name is not None:
            # TODO: remove when getting the exception is implemented
            self._log_error(
                CompilerError.NotSupportedOperation(line=node.lineno,
                                                    col=node.col_offset,
                                                    symbol_id='naming exceptions')
            )
            logged_errors = True

        if not logged_errors and node.type is not None:
            self._log_using_specific_exception_warning(node.type)

        self.generic_visit(node)

    def _log_using_specific_exception_warning(self, node: ast.AST):
        if node is None:
            exc_id = 'Exception'
        elif isinstance(node, ast.Name):
            exc_id = node.id
        else:
            exc_id = self.get_type(node).identifier

        self._log_warning(
            CompilerWarning.UsingSpecificException(line=node.lineno,
                                                   col=node.col_offset,
                                                   exception_id=exc_id)
        )

    def visit_value(self, node: ast.AST):
        result = self.visit(node)

        if isinstance(node, ast.Attribute):
            if isinstance(result, str):
                return self.get_symbol(result)
            else:
                origin, value, attribute = result.values
                if value is None and isinstance(origin, ast.Name):
                    value = self.get_symbol(origin.id)
                if value is not None:
                    return value

        return result

    def visit_Attribute(self, attribute: ast.Attribute) -> Union[str, Attribute]:
        """
        Gets the attribute inside the ast node

        :param attribute: the python ast attribute node
        :return: returns the type of the value, the attribute symbol and its id if the attribute exists.
                 Otherwise, returns None
        """
        is_internal = hasattr(attribute, 'is_internal_call') and attribute.is_internal_call
        value: Optional[Union[str, ISymbol]] = self.get_symbol(attribute.value.id, is_internal) \
            if isinstance(attribute.value, ast.Name) else self.visit(attribute.value)

        if value is None and isinstance(attribute.value, ast.Name):
            return '{0}.{1}'.format(attribute.value.id, attribute.attr)

        symbol = None
        if isinstance(value, str) and not isinstance(attribute.value, ast.Str):
            symbol = self.get_symbol(value)
            if symbol is None:
                return '{0}.{1}'.format(value, attribute.attr)
        if isinstance(value, ISymbol):
            symbol = value

        if isinstance(symbol, Attribute):
            symbol = symbol.attr_symbol
        if isinstance(symbol, IExpression):
            symbol = symbol.type
        if hasattr(symbol, 'symbols') and attribute.attr in symbol.symbols:
            attr_symbol = symbol.symbols[attribute.attr]
        elif isinstance(symbol, Package) and attribute.attr in symbol.inner_packages:
            attr_symbol = symbol.inner_packages[attribute.attr]
        else:
            attr_symbol: Optional[ISymbol] = self.get_symbol(attribute.attr)

        origin = value
        module_symbols = origin.symbols if isinstance(origin, Package) else symbol.methods if isinstance(symbol, Import) else None
        if isinstance(origin, Attribute):
            while isinstance(origin, Attribute):
                origin = origin.value
        else:
            origin = attribute.value

        is_invalid_method = module_symbols is not None and isinstance(attr_symbol, Method) and attribute.attr not in module_symbols
        is_from_class_name = isinstance(origin, ast.Name) and isinstance(self.get_symbol(origin.id), UserClass)
        is_instance_variable_from_class = (isinstance(symbol, UserClass)
                                           and attribute.attr in symbol.instance_variables)
        is_class_variable_from_class = isinstance(symbol, UserClass) and attribute.attr in symbol.class_variables
        is_property_from_class = isinstance(symbol, UserClass) and attribute.attr in symbol.properties

        if ((attr_symbol is None and hasattr(symbol, 'symbols'))
                or is_invalid_method
                or (is_from_class_name and is_instance_variable_from_class)
                or (is_from_class_name and is_property_from_class)):
            # if it couldn't find the symbol in the attribute symbols, raise unresolved reference
            self._log_error(
                CompilerError.UnresolvedReference(
                    attribute.lineno, attribute.col_offset,
                    symbol_id='{0}.{1}'.format(symbol.identifier if module_symbols is None else attribute.value.id, attribute.attr)
                ))
            return Attribute(attribute.value, None, attr_symbol, attribute)

        if not is_from_class_name and is_class_variable_from_class and isinstance(attribute.ctx, ast.Store):
            # reassign class variables in objects is not supported yet
            self._log_error(
                CompilerError.NotSupportedOperation(
                    attribute.lineno, attribute.col_offset,
                    symbol_id='reassign class variables'
                ))
            return Attribute(attribute.value, None, attr_symbol, attribute)

        if not is_from_class_name and is_property_from_class and isinstance(attribute.ctx, ast.Store):
            # setting values for properties in objects is not supported yet
            # @property.setter is not implemented
            self._log_error(
                CompilerError.NotSupportedOperation(
                    attribute.lineno, attribute.col_offset,
                    symbol_id='setting values for properties'
                ))
            return Attribute(attribute.value, None, attr_symbol, attribute)

        attr_type = value.type if isinstance(value, IExpression) else value
        # for checking during the code generation
        if (self.is_implemented_class_type(attr_type) and
                not (isinstance(attribute.value, ast.Name) and attribute.value.id == attr_type.identifier) and
                (not hasattr(attribute, 'generate_value') or not attribute.generate_value)):
            attribute.generate_value = True

        if (isinstance(symbol, (Package, Attribute))
                or (isinstance(symbol, ClassType) and isinstance(value, (Package, Attribute)))):
            attr_value = symbol if not isinstance(symbol, PythonClass) else attribute.value
        else:
            attr_value = attribute.value

        if isinstance(attr_symbol, Property):
            if isinstance(attribute.ctx, ast.Load):
                attr_symbol.getter.add_call_origin(attribute)
        return Attribute(attr_value, attribute.attr, attr_symbol, attribute)

    def visit_Constant(self, constant: ast.Constant) -> Any:
        """
        Visitor of constant values node

        :param constant: the python ast constant value node
        :return: the value of the constant
        """
        if isinstance(constant, ast.Num) and not isinstance(constant.value, int):
            # only integer numbers are allowed
            self._log_error(
                CompilerError.InvalidType(constant.lineno, constant.col_offset, symbol_id=type(constant.value).__name__)
            )
        return constant.value

    def visit_Num(self, num: ast.Num) -> int:
        """
        Verifies if the number is an integer

        :param num: the python ast number node
        :return: returns the value of the number
        """
        if not isinstance(num.n, int):
            # only integer numbers are allowed
            self._log_error(
                CompilerError.InvalidType(num.lineno, num.col_offset, symbol_id=type(num.n).__name__)
            )
        return num.n

    def visit_Str(self, string: ast.Str) -> str:
        """
        Visitor of literal string node

        :param string: the python ast string node
        :return: the value of the string
        """
        return string.s

    def visit_Bytes(self, bts: ast.Bytes) -> bytes:
        """
        Visitor of literal bytes node

        :param bts: the python ast bytes node
        :return: the value of the bytes
        """
        return bts.s

    def visit_Tuple(self, tup_node: ast.Tuple) -> Tuple[Any, ...]:
        """
        Visitor of literal tuple node

        :param tup_node: the python ast tuple node
        :return: the value of the tuple
        """
        return tuple(self.get_type(value) for value in tup_node.elts)

    def visit_List(self, list_node: ast.List) -> List[Any]:
        """
        Visitor of literal list node

        :param list_node: the python ast list node
        :return: the value of the list
        """
        return [self.get_type(value) for value in list_node.elts]

    def visit_Dict(self, dict_node: ast.Dict) -> Dict[Any, Any]:
        """
        Visitor of literal dict node

        :param dict_node: the python ast dict node
        :return: a list with each key and value type
        """
        dictionary = {}
        size = min(len(dict_node.keys), len(dict_node.values))
        for index in range(size):
            key = self.get_type(dict_node.keys[index])
            value = self.get_type(dict_node.values[index])
            if key in dictionary and dictionary[key] != value:
                dictionary[key] = Type.get_generic_type(dictionary[key], value)
            else:
                dictionary[key] = value
        return dictionary

    def visit_NameConstant(self, constant: ast.NameConstant) -> Any:
        """
        Visitor of constant names node

        :param constant: the python ast name constant node
        :return: the value of the constant
        """
        return constant.value

    def visit_Name(self, name: ast.Name) -> ast.Name:
        """
        Visitor of a name node

        :param name:
        :return: the object with the name node information
        """
        return name

    def visit_Starred(self, node: ast.Starred) -> ast.AST:
        value_type = self.get_type(node.value)
        if not Type.sequence.is_type_of(value_type):
            self._log_error(
                CompilerError.MismatchedTypes(line=node.lineno, col=node.col_offset,
                                              expected_type_id=Type.sequence.identifier,
                                              actual_type_id=value_type.identifier)
            )

        return Type.tuple.build_collection(value_type.value_type)

    def visit_Index(self, index: ast.Index) -> Any:
        """
        Visitor of an index node

        :param index:
        :return: the object with the index value information
        """
        index = self.visit(index.value)
        if isinstance(index, (Iterable, Tuple)):
            return tuple(index)
        return index

    def visit_Slice(self, slice_node: ast.Slice) -> Tuple[Any, Any, Any]:
        """
        Visitor of an slice node

        :param slice_node:
        :return: the object with the index value information
        """
        return slice_node.lower, slice_node.upper, slice_node.step



from __future__ import annotations

import ast
from typing import Optional, Union

from boa3.model.expression import IExpression
from boa3.model.type.itype import IType


class Variable(IExpression):
    """
    A class used to represent a variable

    :ivar var_type: the type of the variable.
    """

    def __init__(self, var_type: Optional[IType], origin_node: Optional[ast.AST] = None):
        super().__init__(origin_node)
        self.defined_by_entry = True
        if var_type is None:
            from boa3.analyser.model.optimizer import Undefined, UndefinedType
            var_type = Undefined

        self._var_type: Union[IType, UndefinedType] = var_type

        self.is_reassigned = False
        self._origin_variable: Optional[Variable] = None

    def copy(self) -> Variable:
        var = Variable(self._var_type, self._origin_node)
        var.is_reassigned = self.is_reassigned
        var._origin_variable = self._origin_variable if self._origin_variable is not None else self
        return var

    @property
    def shadowing_name(self) -> str:
        return 'variable'

    @property
    def type(self) -> IType:
        return self._var_type

    def __repr__(self) -> str:
        return '{0}({1})'.format(self.__class__.__name__, self.__str__())

    def __str__(self) -> str:
        return str(self.type)

    def set_type(self, var_type: IType):
        """
        Sets a type for the variable

        :param var_type:
        """
        self._var_type = var_type

    def set_is_reassigned(self):
        if not self.is_reassigned:
            self.is_reassigned = True
        if hasattr(self._origin_variable, 'set_is_reassigned'):
            self._origin_variable.set_is_reassigned()



from typing import Dict

from boa3.model.builtin.contract import NeoAccountStateType
from boa3.model.builtin.interop.nativecontract import NeoContractMethod
from boa3.model.variable import Variable


class GetAccountStateMethod(NeoContractMethod):

    def __init__(self, neo_account_state: NeoAccountStateType):
        from boa3.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'get_account_state'
        native_identifier = 'getAccountState'
        args: Dict[str, Variable] = {
            'account': Variable(UInt160Type.build())
        }
        super().__init__(identifier, native_identifier, args, return_type=neo_account_state)



import ast
from typing import Dict

from boa3.model import set_internal_call
from boa3.model.builtin.interop.nativecontract import StdLibMethod
from boa3.model.variable import Variable


class MemorySearchMethod(StdLibMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.primitive.bytestringtype import ByteStringType

        identifier = 'memory_search'
        native_identifier = 'memorySearch'
        byte_string_type = ByteStringType.build()

        args: Dict[str, Variable] = {
            'mem': Variable(byte_string_type),
            'value': Variable(byte_string_type),
            'start': Variable(Type.int),
            'backward': Variable(Type.bool),
        }

        start_default = set_internal_call(ast.parse("{0}".format(Type.int.default_value)
                                                    ).body[0].value)
        backward_default = set_internal_call(ast.parse("{0}".format(Type.bool.default_value)
                                                       ).body[0].value)

        super().__init__(identifier, native_identifier, args,
                         defaults=[start_default, backward_default],
                         return_type=Type.int)



from typing import Dict

from boa3.model.builtin.interop.nativecontract import NeoContractMethod
from boa3.model.variable import Variable


class UnregisterCandidateMethod(NeoContractMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.collection.sequence.ecpointtype import ECPointType

        identifier = 'unregister_candidate'
        native_identifier = 'unregisterCandidate'
        args: Dict[str, Variable] = {
            'pubkey': Variable(ECPointType.build())
        }
        super().__init__(identifier, native_identifier, args, return_type=Type.bool)



from typing import Any

from boa3.neo.vm.type import StackItem
from boa3.neo.vm.type.StackItem import StackItemType


class NativeAccountState:
    def __init__(self, balance: int):
        self.balance: int = balance if balance > 0 else 0
        self._balance_height: int = 0
        self._vote_to: Any = None

    def serialize(self) -> bytes:
        # TODO: change this when Struct is implemented
        return (StackItemType.Struct
                + b'\x03'
                + StackItem.serialize(self.balance)
                + b'!\x00\x00'
                )



__all__ = ['BooleanAnd',
           'BooleanOr',
           'LeftShift',
           'LogicAnd',
           'LogicOr',
           'LogicXor',
           'RightShift'
           ]

from boa3.model.operation.binary.logical.booleanand import BooleanAnd
from boa3.model.operation.binary.logical.booleanor import BooleanOr
from boa3.model.operation.binary.logical.leftshift import LeftShift
from boa3.model.operation.binary.logical.logicand import LogicAnd
from boa3.model.operation.binary.logical.logicor import LogicOr
from boa3.model.operation.binary.logical.logicxor import LogicXor
from boa3.model.operation.binary.logical.rightshift import RightShift



from __future__ import annotations

from typing import Any, Dict, Optional

from boa3.neo3.core.types import UInt160
from boa3_test.tests.test_classes.contract.neostruct import NeoStruct


class NeoPermissionsStruct(NeoStruct):
    _contract_fields = 'contract'
    _methods_fields = 'methods'

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> NeoPermissionsStruct:
        required_fields = [cls._contract_fields,
                           cls._methods_fields
                           ]
        cls._validate_json(json, required_fields)

        struct = cls()
        struct.append(cls.get_contract(json[cls._contract_fields]))
        struct.append(cls.get_methods(json[cls._methods_fields]))
        return struct

    @classmethod
    def _is_wildcard(cls, value: Any) -> bool:
        if value is None:
            return True
        return value == '*'

    @classmethod
    def get_contract(cls, value: Any) -> Optional[UInt160]:
        if cls._is_wildcard(value):
            return None

        # TODO: Permissions are not implemented yet
        return UInt160()

    @classmethod
    def get_methods(cls, value: Any) -> Optional[list]:
        if cls._is_wildcard(value):
            return None

        # TODO: Permissions are not implemented yet
        return []



from __future__ import annotations

from typing import Any, Dict

from boa3.neo import to_hex_str
from boa3.neo3.core.types import UInt160


class TestRunnerLog:
    _contract_key = 'contract'
    _message_key = 'message'

    def __init__(self, script_hash: bytes, message: str):
        self._script_hash: bytes = script_hash
        self._message: str = message

    @property
    def origin(self) -> bytes:
        return self._script_hash

    @property
    def message(self) -> str:
        return self._message

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> TestRunnerLog:
        """
        Creates a Log object from a json.

        :param json: json that contains the log data
        :return: a Log object
        :rtype: TestRunnerLog
        """
        keys = set(json.keys())
        if not keys.issubset([cls._contract_key, cls._message_key]):
            return None

        script: bytes = json[cls._contract_key] if cls._contract_key in json else b''
        message: str = json[cls._message_key] if cls._message_key in json else ""

        if isinstance(script, str):
            script = UInt160.from_string(script[2:]
                                         if script.startswith('0x')
                                         else script
                                         ).to_array()

        return cls(script, message)

    def __str__(self) -> str:
        return '[{0}] {1}'.format(to_hex_str(self.origin), self._message)

    def __repr__(self) -> str:
        return str(self)
